# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re

from . import SectionNotFoundError


class ConfigFileWriter:
    SECTION_REGEX = re.compile(r'^\s*\[(?P<header>[^]]+)\]')
    OPTION_REGEX = re.compile(
        r'(?P<option>[^:=][^:=]*)'
        r'\s*(?P<vi>[:=])\s*'
        r'(?P<value>.*)$'
    )  # fmt: skip

    def _validate_no_newlines_or_carriage_returns(
        self,
        value,
        label='value',
        msg_override=None,
    ):
        if isinstance(value, str) and ('\n' in value or '\r' in value):
            err_msg = msg_override if msg_override is not None else (
                f"Invalid {label}: newline "
                f"characters and carriage returns are not allowed: {value!r}"
            )
            raise ValueError(err_msg)

    def update_config(self, new_values, config_filename):
        """Update config file with new values.

        This method will update a section in a config file with
        new key value pairs.

        This method provides a few conveniences:

        * If the ``config_filename`` does not exist, it will
          be created.  Any parent directories will also be created
          if necessary.
        * If the section to update does not exist, it will be created.
        * Any existing lines that are specified by ``new_values``
          **will not be touched**.  This ensures that commented out
          values are left unaltered.

        :type new_values: dict
        :param new_values: The values to update.  There is a special
            key ``__section__``, that specifies what section in the INI
            file to update.  If this key is not present, then the
            ``default`` section will be updated with the new values.

        :type config_filename: str
        :param config_filename: The config filename where values will be
            written.

        """
        section_name = new_values.pop('__section__', 'default')
        self._validate_no_newlines_or_carriage_returns(
            section_name,
            'section name'
        )
        for k, v in new_values.items():
            self._validate_no_newlines_or_carriage_returns(k, 'key')
            if not isinstance(v, dict):
                # Override error msg to prevent
                # leaking sensitive config values to stderr.
                self._validate_no_newlines_or_carriage_returns(
                    v,
                    'value',
                    msg_override=(
                        f"Invalid value for key {k}: "
                        f"newline characters and carriage "
                        f"returns are not allowed."
                    )
                )
            else:
                for sk, sv in v.items():
                    # Override error msg to prevent
                    # leaking sensitive config values to stderr.
                    self._validate_no_newlines_or_carriage_returns(sk, 'key')
                    self._validate_no_newlines_or_carriage_returns(
                        sv,
                        'value',
                        msg_override=(
                            f"Invalid value for key {k}: "
                            f"newline characters and carriage "
                            f"returns are not allowed."
                        )
                    )
        if not os.path.isfile(config_filename):
            self._create_file(config_filename)
            self._write_new_section(section_name, new_values, config_filename)
            return
        with open(config_filename) as f:
            contents = f.readlines()
        # We can only update a single section at a time so we first need
        # to find the section in question
        try:
            self._update_section_contents(contents, section_name, new_values)
            with open(config_filename, 'w') as f:
                f.write(''.join(contents))
        except SectionNotFoundError:
            self._write_new_section(section_name, new_values, config_filename)

    def _create_file(self, config_filename):
        # Create the file as well as the parent dir if needed.
        dirname = os.path.split(config_filename)[0]
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with os.fdopen(
            os.open(config_filename, os.O_WRONLY | os.O_CREAT, 0o600), 'w'
        ):
            pass

    def _check_file_needs_newline(self, filename):
        # check if the last byte is a newline
        with open(filename, 'rb') as f:
            # check if the file is empty
            f.seek(0, os.SEEK_END)
            if not f.tell():
                return False
            f.seek(-1, os.SEEK_END)
            last = f.read()
            return last != b'\n'

    def _write_new_section(self, section_name, new_values, config_filename):
        needs_newline = self._check_file_needs_newline(config_filename)
        with open(config_filename, 'a') as f:
            if needs_newline:
                f.write('\n')
            f.write(f'[{section_name}]\n')
            contents = []
            self._insert_new_values(
                line_number=0, contents=contents, new_values=new_values
            )
            f.write(''.join(contents))

    def _find_section_start(self, contents, section_name):
        for i in range(len(contents)):
            line = contents[i]
            if line.strip().startswith(('#', ';')):
                # This is a comment, so we can safely ignore this line.
                continue
            match = self.SECTION_REGEX.search(line)
            if match is not None and self._matches_section(
                match, section_name
            ):
                return i
        raise SectionNotFoundError(section_name)

    def _update_section_contents(self, contents, section_name, new_values):
        # First, find the line where the section_name is defined.
        # This will be the value of i.
        new_values = new_values.copy()
        # ``contents`` is a list of file line contents.
        section_start_line_num = self._find_section_start(
            contents, section_name
        )
        # If we get here, then we've found the section.  We now need
        # to figure out if we're updating a value or adding a new value.
        # There's 2 cases.  Either we're setting a normal scalar value
        # of, we're setting a nested value.
        last_matching_line = section_start_line_num
        j = last_matching_line + 1
        while j < len(contents):
            line = contents[j]
            if self.SECTION_REGEX.search(line) is not None:
                # We've hit a new section which means the config key is
                # not in the section.  We need to add it here.
                self._insert_new_values(
                    line_number=last_matching_line,
                    contents=contents,
                    new_values=new_values,
                )
                return
            match = self.OPTION_REGEX.search(line)
            if match is not None:
                last_matching_line = j
                key_name = match.group(1).strip()
                if key_name in new_values:
                    # We've found the line that defines the option name.
                    # if the value is not a dict, then we can write the line
                    # out now.
                    if not isinstance(new_values[key_name], dict):
                        option_value = new_values[key_name]
                        if option_value is None:
                            # Remove the line by replacing with empty string
                            contents[j] = ''
                        else:
                            new_line = f'{key_name} = {option_value}\n'
                            contents[j] = new_line
                        del new_values[key_name]
                    else:
                        j = self._update_subattributes(
                            j,
                            contents,
                            new_values[key_name],
                            len(match.group(1)) - len(match.group(1).lstrip()),
                        )
                        return
            j += 1

        if new_values:
            if not contents[-1].endswith('\n'):
                contents.append('\n')
            self._insert_new_values(
                line_number=last_matching_line + 1,
                contents=contents,
                new_values=new_values,
            )

    def _update_subattributes(self, index, contents, values, starting_indent):
        index += 1
        for i in range(index, len(contents)):
            line = contents[i]
            match = self.OPTION_REGEX.search(line)
            if match is not None:
                current_indent = len(match.group(1)) - len(
                    match.group(1).lstrip()
                )
                key_name = match.group(1).strip()
                if key_name in values:
                    option_value = values[key_name]
                    new_line = (
                        f"{' ' * current_indent}{key_name} = {option_value}\n"
                    )
                    contents[i] = new_line
                    del values[key_name]
            if (
                starting_indent == current_indent
                or self.SECTION_REGEX.search(line) is not None
            ):
                # We've arrived at the starting indent level so we can just
                # write out all the values now.
                self._insert_new_values(i - 1, contents, values, '    ')
                break
        else:
            if starting_indent != current_indent:
                # The option is the last option in the file
                self._insert_new_values(i, contents, values, '    ')
        return i

    def _insert_new_values(self, line_number, contents, new_values, indent=''):
        new_contents = []
        for key, value in list(new_values.items()):
            if isinstance(value, dict):
                subindent = indent + '    '
                new_contents.append(f'{indent}{key} =\n')
                for subkey, subval in list(value.items()):
                    new_contents.append(f'{subindent}{subkey} = {subval}\n')
            elif value is not None:
                new_contents.append(f'{indent}{key} = {value}\n')
            del new_values[key]
        if new_contents:
            contents.insert(line_number + 1, ''.join(new_contents))

    def _matches_section(self, match, section_name):
        parts = section_name.split(' ')
        unquoted_match = match.group(0) == f'[{section_name}]'
        if len(parts) > 1:
            quoted_match = (
                match.group(0) == f'[{parts[0]} "{" ".join(parts[1:])}"]'
            )
            return unquoted_match or quoted_match
        return unquoted_match



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# Declare all the constants used by Lifecycle in this file

# Lifecycle role names
LIFECYCLE_DEFAULT_ROLE_NAME = "AWSDataLifecycleManagerDefaultRole"
LIFECYCLE_DEFAULT_ROLE_NAME_AMI = (
    "AWSDataLifecycleManagerDefaultRoleForAMIManagement"
)

# Lifecycle role arn names
LIFECYCLE_DEFAULT_MANAGED_POLICY_NAME = "AWSDataLifecycleManagerServiceRole"
LIFECYCLE_DEFAULT_MANAGED_POLICY_NAME_AMI = (
    "AWSDataLifecycleManagerServiceRoleForAMIManagement"
)

POLICY_ARN_PATTERN = "arn:{0}:iam::aws:policy/service-role/{1}"

# Assume Role Policy definitions for roles
LIFECYCLE_DEFAULT_ROLE_ASSUME_POLICY = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "",
            "Effect": "Allow",
            "Principal": {"Service": "dlm.amazonaws.com"},
            "Action": "sts:AssumeRole",
        }
    ],
}

RESOURCE_TYPE_SNAPSHOT = "snapshot"
RESOURCE_TYPE_IMAGE = "image"

RESOURCES = {
    RESOURCE_TYPE_SNAPSHOT: {
        'default_role_name': LIFECYCLE_DEFAULT_ROLE_NAME,
        'default_policy_name': LIFECYCLE_DEFAULT_MANAGED_POLICY_NAME,
    },
    RESOURCE_TYPE_IMAGE: {
        'default_role_name': LIFECYCLE_DEFAULT_ROLE_NAME_AMI,
        'default_policy_name': LIFECYCLE_DEFAULT_MANAGED_POLICY_NAME_AMI,
    },
}



"""Placeholder module for fleshing out the server side model interface."""

import json


class DBCompletionLookup:
    _QUERY = """\
        SELECT apicall_data from apicall_table
        INNER JOIN param_table
        ON apicall_table.param_id = param_table.param_id
        WHERE
          argname = :argname AND
          command = :command AND
          parent = :parent
    """

    def __init__(self, db_connection):
        self._db_connection = db_connection

    def get_server_completion_data(self, lineage, command_name, param_name):
        parent = '.'.join(lineage)
        results = self._db_connection.execute(
            self._QUERY,
            argname=param_name,
            command=command_name,
            parent=parent,
        )
        match = results.fetchone()
        if match is not None:
            return json.loads(match[0])



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
Utility functions to make it easier to work with customizations.

"""

import copy
import re
import sys
import xml

from botocore.exceptions import ClientError

from awscli.customizations.exceptions import ParamValidationError

_SENTENCE_DELIMETERS_REGEX = re.compile(r'[.:]+')
_LINE_BREAK_CHARS = ['\n', '\u2028']


def rename_argument(argument_table, existing_name, new_name):
    current = argument_table[existing_name]
    argument_table[new_name] = current
    current.name = new_name
    del argument_table[existing_name]


def _copy_argument(argument_table, current_name, copy_name):
    current = argument_table[current_name]
    copy_arg = copy.copy(current)
    copy_arg.name = copy_name
    argument_table[copy_name] = copy_arg
    return copy_arg


def make_hidden_alias(argument_table, existing_name, alias_name):
    """Create a hidden alias for an existing argument.
    This will copy an existing argument object in an arg table,
    and add a new entry to the arg table with a different name.
    The new argument will also be undocumented.
    This is needed if you want to check an existing argument,
    but you still need the other one to work for backwards
    compatibility reasons.
    """
    current = argument_table[existing_name]
    copy_arg = _copy_argument(argument_table, existing_name, alias_name)
    copy_arg._UNDOCUMENTED = True
    if current.required:
        # If the current argument is required, then
        # we'll mark both as not required, but
        # flag _DOCUMENT_AS_REQUIRED so our doc gen
        # knows to still document this argument as required.
        copy_arg.required = False
        current.required = False
        current._DOCUMENT_AS_REQUIRED = True


def rename_command(command_table, existing_name, new_name):
    current = command_table[existing_name]
    command_table[new_name] = current
    current.name = new_name
    del command_table[existing_name]


def alias_command(command_table, existing_name, new_name):
    """Moves an argument to a new name, keeping the old as a hidden alias.

    :type command_table: dict
    :param command_table: The full command table for the CLI or a service.

    :type existing_name: str
    :param existing_name: The current name of the command.

    :type new_name: str
    :param new_name: The new name for the command.
    """
    current = command_table[existing_name]
    _copy_argument(command_table, existing_name, new_name)
    current._UNDOCUMENTED = True


def validate_mutually_exclusive_handler(*groups):
    def _handler(parsed_args, **kwargs):
        return validate_mutually_exclusive(parsed_args, *groups)

    return _handler


def validate_mutually_exclusive(parsed_args, *groups):
    """Validate mutually exclusive groups in the parsed args."""
    args_dict = vars(parsed_args)
    all_args = set(arg for group in groups for arg in group)
    if not any(k in all_args for k in args_dict if args_dict[k] is not None):
        # If none of the specified args are in a mutually exclusive group
        # there is nothing left to validate.
        return
    current_group = None
    for key in [k for k in args_dict if args_dict[k] is not None]:
        key_group = _get_group_for_key(key, groups)
        if key_group is None:
            # If they key is not part of a mutex group, we can move on.
            continue
        if current_group is None:
            current_group = key_group
        elif not key_group == current_group:
            raise ParamValidationError(
                'The key "%s" cannot be specified when one '
                'of the following keys are also specified: '
                '%s' % (key, ', '.join(current_group))
            )


def _get_group_for_key(key, groups):
    for group in groups:
        if key in group:
            return group


def s3_bucket_exists(s3_client, bucket_name):
    bucket_exists = True
    try:
        # See if the bucket exists by running a head bucket
        s3_client.head_bucket(Bucket=bucket_name)
    except ClientError as e:
        # If a client error is thrown. Check that it was a 404 error.
        # If it was a 404 error, than the bucket does not exist.
        error_code = int(e.response['Error']['Code'])
        if error_code == 404:
            bucket_exists = False
    return bucket_exists


def create_client_from_parsed_globals(
    session, service_name, parsed_globals, overrides=None
):
    """Creates a service client, taking parsed_globals into account

    Any values specified in overrides will override the returned dict. Note
    that this override occurs after 'region' from parsed_globals has been
    translated into 'region_name' in the resulting dict.
    """
    client_args = {}
    if 'region' in parsed_globals:
        client_args['region_name'] = parsed_globals.region
    if 'endpoint_url' in parsed_globals:
        client_args['endpoint_url'] = parsed_globals.endpoint_url
    if 'verify_ssl' in parsed_globals:
        client_args['verify'] = parsed_globals.verify_ssl
    if overrides:
        client_args.update(overrides)
    return session.create_client(service_name, **client_args)


def uni_print(statement, out_file=None):
    """
    This function is used to properly write unicode to a file, usually
    stdout or stdderr.  It ensures that the proper encoding is used if the
    statement is not a string type.
    """
    if out_file is None:
        out_file = sys.stdout
    try:
        # Otherwise we assume that out_file is a
        # text writer type that accepts str/unicode instead
        # of bytes.
        out_file.write(statement)
    except UnicodeEncodeError:
        # Some file like objects like cStringIO will
        # try to decode as ascii on python2.
        #
        # This can also fail if our encoding associated
        # with the text writer cannot encode the unicode
        # ``statement`` we've been given.  This commonly
        # happens on windows where we have some S3 key
        # previously encoded with utf-8 that can't be
        # encoded using whatever codepage the user has
        # configured in their console.
        #
        # At this point we've already failed to do what's
        # been requested.  We now try to make a best effort
        # attempt at printing the statement to the outfile.
        # We're using 'ascii' as the default because if the
        # stream doesn't give us any encoding information
        # we want to pick an encoding that has the highest
        # chance of printing successfully.
        new_encoding = getattr(out_file, 'encoding', 'ascii')
        # When the output of the aws command is being piped,
        # ``sys.stdout.encoding`` is ``None``.
        if new_encoding is None:
            new_encoding = 'ascii'
        new_statement = statement.encode(new_encoding, 'replace').decode(
            new_encoding
        )
        out_file.write(new_statement)
    out_file.flush()


def get_policy_arn_suffix(region):
    """Method to return region value as expected by policy arn"""
    region_string = region.lower()
    if region_string.startswith("cn-"):
        return "aws-cn"
    elif region_string.startswith("us-gov"):
        return "aws-us-gov"
    else:
        return "aws"


def get_shape_doc_overview(shape):
    """Return a documentation overview of a shape

    Currently, this just returns the first sentence of the shape's
    documentation.
    """
    content = shape.documentation
    content = _strip_xml_from_documentation(content)
    # In order to avoid having the content too dense, we limit
    # the documentation to the first sentence.
    content = _get_first_sentence(content)
    # There are characters that may mess up the indentation by introducing new
    # lines. We want to ignore those for the purpose of an overview.
    content = _remove_line_breaks(content)
    return content


def _strip_xml_from_documentation(documentation):
    try:
        # We are surrounding the docstrings with our own tags in order
        # to make sure the dom parser will look at all elements in the
        # docstring as some docstrings may not have xml nodes that do
        # not all belong to the same root node.
        xml_doc = '<doc>%s</doc>' % documentation
        xml_dom = xml.dom.minidom.parseString(xml_doc)
    except xml.parsers.expat.ExpatError:
        return documentation
    content = []
    _strip_xml_from_child_nodes(xml_dom, content)
    return ''.join(content)


def _strip_xml_from_child_nodes(node, content):
    for child_node in node.childNodes:
        if child_node.nodeType == node.TEXT_NODE:
            content.append(child_node.data)
        else:
            _strip_xml_from_child_nodes(child_node, content)


def _get_first_sentence(content):
    content = _SENTENCE_DELIMETERS_REGEX.split(content, 1)[0]
    if content:
        content += '.'
    return content


def _remove_line_breaks(content):
    for char in _LINE_BREAK_CHARS:
        content = content.replace(char, ' ')
    return content



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys
from datetime import datetime, timezone

import dateutil.parser

from awscli.customizations.ecs.expressgateway.color_utils import ColorUtils

MANAGED_RESOURCE_REASON_MESSAGE = "{indent}Reason: {reason}"
MANAGED_RESOURCE_LAST_UPDATED_MESSAGE = (
    "{indent}Last updated at: {last_updated_at}"
)

TERMINAL_RESOURCE_STATUSES = ["ACTIVE", "DELETED", "FAILED"]
TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class ManagedResource:
    """
    Represents a managed ECS resource.
    """

    def __init__(
        self,
        resource_type,
        identifier,
        status=None,
        updated_at=None,
        reason=None,
        additional_info=None,
    ):
        self.resource_type = resource_type
        self.identifier = identifier
        self.status = status
        if isinstance(updated_at, str):
            dt = dateutil.parser.parse(updated_at)
            self.updated_at = dt.timestamp()
        else:
            self.updated_at = updated_at
        self.reason = reason
        self.additional_info = additional_info
        self.color_utils = ColorUtils()

    def is_terminal(self):
        return self.status in TERMINAL_RESOURCE_STATUSES

    def get_status_string(self, spinner_char, depth=0, use_color=True):
        """Returns the resource information as a formatted string.

        Args:
            spinner_char (str): Character to display for in-progress resources
            depth (int): Indentation depth for nested display (default: 0)
            use_color (bool): Whether to use ANSI color codes (default: True)

        Returns:
            str: Formatted status string with resource information
        """
        lines = []
        resource_header = " " * depth + self.color_utils.make_cyan(
            self.resource_type, use_color
        )

        resource_header += ": " if self.identifier else " "
        resource_header += self.color_utils.make_status_symbol(
            self.status, spinner_char, use_color
        )
        if self.identifier:
            resource_header += (
                self.color_utils.color_by_status(
                    self.identifier, self.status, use_color
                )
                + " "
            )
        if self.status:
            resource_header += "- " + self.color_utils.color_by_status(
                self.status, self.status, use_color
            )
        lines.append(resource_header)

        if self.reason:
            lines.append(
                MANAGED_RESOURCE_REASON_MESSAGE.format(
                    indent=" " * (depth + 1), reason=self.reason
                )
            )

        if self.updated_at:
            lines.append(
                MANAGED_RESOURCE_LAST_UPDATED_MESSAGE.format(
                    indent=" " * (depth + 1),
                    last_updated_at=datetime.fromtimestamp(
                        self.updated_at
                    ).strftime(TIMESTAMP_FORMAT),
                )
            )

        if self.additional_info:
            lines.append(" " * (depth + 1) + self.additional_info)

        # Spacing between resources
        lines.append("")
        return '\n'.join(lines)

    def get_stream_string(self, timestamp, use_color=True):
        """Returns the resource information formatted for stream/text-only display.

        Args:
            timestamp (str): Timestamp string to prefix the output
            use_color (bool): Whether to use ANSI color codes (default: True)

        Returns:
            str: Formatted string with timestamp prefix and bracket-enclosed status
        """
        lines = []
        parts = [f"[{timestamp}]"]

        # If both resource_type and identifier are None, show a placeholder
        if not self.resource_type and not self.identifier:
            parts.append(
                self.color_utils.make_cyan("Unknown Resource", use_color)
            )
        else:
            if self.resource_type:
                parts.append(
                    self.color_utils.make_cyan(self.resource_type, use_color)
                )

            if self.identifier:
                colored_id = self.color_utils.color_by_status(
                    self.identifier, self.status, use_color
                )
                parts.append(colored_id)

        if self.status:
            status_text = self.color_utils.color_by_status(
                self.status, self.status, use_color
            )
            parts.append(f"[{status_text}]")

        lines.append(" ".join(parts))

        if self.reason:
            lines.append(f"  Reason: {self.reason}")

        if self.updated_at:
            updated_time = datetime.fromtimestamp(
                self.updated_at, tz=timezone.utc
            ).strftime("%Y-%m-%d %H:%M:%SZ")
            lines.append(f"  Last Updated At: {updated_time}")

        if self.additional_info:
            lines.append(f"  Info: {self.additional_info}")

        return "\n".join(lines)

    def combine(self, other_resource):
        """Returns the version of the resource which has the most up to date timestamp.

        Args:
            other_resource (ManagedResource): Resource to compare timestamps with

        Returns:
            ManagedResource: The resource with the latest timestamp
        """
        return (
            self
            if not other_resource
            or not other_resource.updated_at
            or self.updated_at >= other_resource.updated_at
            else other_resource
        )

    def compare_properties(self, other_resource):
        """Compares individual resource properties to detect changes.

        This compares properties like status, reason, updated_at, additional_info
        to detect if a resource has changed between polls.

        Args:
            other_resource (ManagedResource): Resource to compare against

        Returns:
            bool: True if properties differ, False if same
        """
        if not other_resource:
            # No previous resource means it's new/different
            return True

        # Resources are different if any field differs
        return (
            self.resource_type != other_resource.resource_type
            or self.identifier != other_resource.identifier
            or self.status != other_resource.status
            or self.reason != other_resource.reason
            or self.updated_at != other_resource.updated_at
            or self.additional_info != other_resource.additional_info
        )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.wizard import core, ui
from awscli.customizations.wizard.app import (
    WizardApp,
    WizardAppRunner,
    WizardTraverser,
    WizardValues,
)
from awscli.customizations.wizard.ui.layout import WizardLayoutFactory


def create_default_executor(api_invoker, shared_config):
    return core.Executor(
        step_handlers={
            core.APICallExecutorStep.NAME: core.APICallExecutorStep(
                api_invoker
            ),
            core.SharedConfigExecutorStep.NAME: core.SharedConfigExecutorStep(
                shared_config
            ),
            core.DefineVariableStep.NAME: core.DefineVariableStep(),
            core.MergeDictStep.NAME: core.MergeDictStep(),
            core.LoadDataExecutorStep.NAME: core.LoadDataExecutorStep(),
            core.DumpDataExecutorStep.NAME: core.DumpDataExecutorStep(),
        }
    )


def create_default_wizard_v1_runner(session):
    api_invoker = core.APIInvoker(session=session)
    shared_config = core.SharedConfigAPI(
        session=session, config_writer=ConfigFileWriter()
    )
    planner = core.Planner(
        step_handlers={
            core.StaticStep.NAME: core.StaticStep(),
            core.PromptStep.NAME: core.PromptStep(ui.UIPrompter()),
            core.YesNoPrompt.NAME: core.YesNoPrompt(ui.UIPrompter()),
            core.FilePromptStep.NAME: core.FilePromptStep(
                ui.UIFilePrompter(ui.FileCompleter())
            ),
            core.TemplateStep.NAME: core.TemplateStep(),
            core.APICallStep.NAME: core.APICallStep(api_invoker=api_invoker),
            core.SharedConfigStep.NAME: core.SharedConfigStep(
                config_api=shared_config
            ),
        }
    )
    executor = create_default_executor(api_invoker, shared_config)
    runner = core.Runner(planner, executor)
    return runner


def create_default_wizard_v2_runner(session):
    return WizardAppRunner(session=session, app_factory=create_wizard_app)


def create_wizard_app(definition, session, output=None, app_input=None):
    api_invoker = core.APIInvoker(session=session)
    shared_config = core.SharedConfigAPI(
        session=session, config_writer=ConfigFileWriter()
    )
    layout = WizardLayoutFactory().create_wizard_layout(definition)
    values = WizardValues(
        definition,
        value_retrieval_steps={
            core.APICallStep.NAME: core.APICallStep(api_invoker=api_invoker),
            core.SharedConfigStep.NAME: core.SharedConfigStep(
                config_api=shared_config
            ),
            core.TemplateStep.NAME: core.TemplateStep(),
            core.LoadDataStep.NAME: core.LoadDataStep(),
            core.DumpDataStep.NAME: core.DumpDataStep(),
        },
        exception_handler=layout.error_bar.display_error,
    )
    executor = create_default_executor(api_invoker, shared_config)
    traverser = WizardTraverser(definition, values, executor)
    return WizardApp(
        layout=layout,
        values=values,
        traverser=traverser,
        executor=executor,
        output=output,
        app_input=app_input,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from decimal import Decimal

import awscli.customizations.dynamodb.ast as ast

from .exceptions import (
    EmptyExpressionError,
    InvalidLiteralValueError,
    UnexpectedTokenError,
    UnknownExpressionError,
)
from .lexer import Lexer
from .types import Binary


class Parser:
    COMPARATORS = ['eq', 'ne', 'lt', 'lte', 'gt', 'gte']

    def __init__(self, lexer=None):
        self._lexer = lexer
        if lexer is None:
            self._lexer = Lexer()
        self._position = 0
        self._tokens = []
        self._current = None
        self._expression = None

    def parse(self, expression):
        self._position = 0
        self._expression = expression
        self._tokens = list(self._lexer.tokenize(expression))
        self._current = self._tokens[0]

        parsed = self._parse_expression()

        if not self._match('eof'):
            raise UnexpectedTokenError(
                token=self._current,
                expression=self._expression,
                expected_type='eof',
            )
        return parsed

    def _parse_expression(self):
        if self._match('eof') or self._current is None:
            raise EmptyExpressionError()

        expression = self._parse_simple_expression()

        identifier_types = [
            'identifier',
            'path_identifier',
            'index_identifier',
        ]
        if expression['type'] in identifier_types and self._match('comma'):
            self._advance()
            return self._parse_sequence(expression)
        else:
            return self._parse_and_or(expression)

    def _parse_and_or(self, left_expression):
        expression = left_expression

        while self._match(['and', 'or']):
            conjunction_type = self._current['type']
            self._advance()
            right = self._parse_simple_expression()
            if conjunction_type == 'and':
                expression = ast.and_expression(expression, right)
            else:
                expression = ast.or_expression(expression, right)

        return expression

    def _parse_simple_expression(self):
        if self._match('lparen'):
            return self._parse_subexpression()
        if self._match('not'):
            return self._parse_not_expression()
        return self._parse_condition_expression()

    def _parse_subexpression(self):
        self._advance_if_match('lparen')
        expression = self._parse_expression()
        self._advance_if_match('rparen')
        return ast.subexpression(expression)

    def _parse_not_expression(self):
        self._advance_if_match('not')
        expression = self._parse_simple_expression()
        return ast.not_expression(expression)

    def _parse_condition_expression(self):
        # function names cannot be literals or complex identifiers, so there
        # is no need to parse the name ahead of time
        if self._match_next('lparen'):
            return self._parse_function()

        operand_types = [
            'literal',
            'identifier',
            'unquoted_identifier',
            'lbracket',
            'lbrace',
        ]
        if not self._match(operand_types):
            raise UnknownExpressionError(
                start_token=self._current,
                expression=self._expression,
            )

        left = self._parse_operand()

        if self._match('in'):
            return self._parse_in_expression(left)
        elif self._match('between'):
            return self._parse_between_expression(left)
        elif self._match(self.COMPARATORS):
            return self._parse_comparison_expression(left)

        return left

    def _parse_function(self):
        function_name = self._current.get('value')
        self._advance_if_match(['identifier', 'unquoted_identifier'])
        self._advance_if_match('lparen')
        arguments = self._parse_sequence()["children"]
        self._advance_if_match('rparen')
        return ast.function_expression(function_name, arguments)

    def _parse_in_expression(self, left):
        self._advance_if_match('in')
        self._advance_if_match('lparen')
        right = self._parse_sequence()
        self._advance_if_match('rparen')
        return ast.in_expression(left, right)

    def _parse_between_expression(self, left):
        self._advance_if_match('between')
        middle = self._parse_operand()
        self._advance_if_match('and')
        right = self._parse_operand()
        return ast.between_expression(left, middle, right)

    def _parse_comparison_expression(self, left):
        comparator = self._current['type']
        self._advance_if_match(self.COMPARATORS)
        right = self._parse_operand()
        return ast.comparison_expression(comparator, left, right)

    def _parse_sequence(self, first_element=None):
        elements = []
        if first_element:
            elements = [first_element]

        while True:
            elements.append(self._parse_operand())
            if not self._match('comma'):
                break
            self._advance()
        return ast.sequence(elements)

    def _parse_operand(self):
        if self._match(['literal', 'lbracket', 'lbrace']):
            return self._parse_literal()
        elif self._match(['identifier', 'unquoted_identifier']):
            return self._parse_path_identifier()
        else:
            raise UnexpectedTokenError(
                token=self._current,
                expression=self._expression,
                expected_type=[
                    'literal',
                    'lbracket',
                    'lbrace',
                    'identifier',
                    'unquoted_identiifer',
                ],
            )

    def _parse_path_identifier(self):
        identifier = self._parse_identifier()

        while self._match('dot'):
            self._advance()
            right = self._parse_identifier()
            identifier = ast.path_identifier(identifier, right)

        return identifier

    def _parse_identifier(self):
        identifier = ast.identifier(self._current['value'])
        self._advance()

        if self._match('lbracket'):
            self._advance()

            index = self._parse_literal()
            val = index['value']
            if not isinstance(val, Decimal) or val.to_integral_value() != val:
                raise InvalidLiteralValueError(
                    token=self._current,
                    expression=self._expression,
                    message='List indices must be whole numbers.',
                )
            self._advance_if_match('rbracket')
            identifier = ast.index_identifier(
                name=identifier, index=index['value']
            )

        return identifier

    def _parse_literal(self):
        if self._match('literal'):
            value = self._current['value']
            self._advance()
        elif self._match('lbracket'):
            self._advance()
            if self._match('rbracket'):
                value = []
            else:
                value = self._parse_literal_sequence()
            self._advance_if_match('rbracket')
        elif self._match('lbrace'):
            self._advance()
            if self._match_next('colon') or self._match('rbrace'):
                value = self._parse_literal_map()
            else:
                value = self._parse_literal_set()
            self._advance_if_match('rbrace')
        else:
            raise UnexpectedTokenError(
                token=self._current,
                expression=self._expression,
                expected_type=['literal', 'lbracket', 'lbrace'],
            )
        return ast.literal(value)

    def _parse_literal_set(self):
        valid_types = (str, Binary, Decimal)
        first_type = type(self._current['value'])

        elements = set()
        while True:
            element = self._current
            if not self._match('literal'):
                raise UnexpectedTokenError(
                    token=self._current,
                    expression=self._expression,
                    expected_type='literal',
                )
            if not isinstance(element['value'], valid_types):
                message = (
                    'Sets may only contain numbers, strings, or bytes, '
                    'but literal of type `%s` was found'
                )
                raise InvalidLiteralValueError(
                    token=self._current,
                    expression=self._expression,
                    message=message % type(element['type']),
                )
            if not isinstance(element['value'], first_type):
                message = (
                    'Set values must all be of the same type. First type was '
                    '`%s`, but found value of type `%s`'
                )
                raise InvalidLiteralValueError(
                    token=self._current,
                    expression=self._expression,
                    message=message % (first_type, type(element['type'])),
                )

            elements.add(self._current['value'])
            self._advance()
            if not self._match('comma'):
                break
            self._advance()
        return elements

    def _parse_literal_map(self):
        if self._match('rbrace'):
            return {}

        elements = {}
        while True:
            key = self._current['value']
            if not isinstance(key, str):
                raise InvalidLiteralValueError(
                    token=self._current,
                    expression=self._expression,
                    message=(
                        'Keys must be of type `str`, found `%s`' % type(key)
                    ),
                )
            self._advance_if_match('literal')
            self._advance_if_match('colon')
            value = self._parse_literal()['value']
            elements[key] = value
            if not self._match('comma'):
                break
            self._advance()
        return elements

    def _parse_literal_sequence(self):
        elements = []
        while True:
            elements.append(self._parse_literal()['value'])
            if not self._match('comma'):
                break
            self._advance()
        return elements

    def _advance(self):
        if self._position == len(self._tokens) - 1:
            self._current = None
        else:
            self._position += 1
            self._current = self._tokens[self._position]

    def _peek(self):
        if self._position == len(self._tokens) - 1:
            return None
        return self._tokens[self._position + 1]

    def _match(self, expected_type):
        return self._do_match(self._current, expected_type)

    def _match_next(self, expected_type):
        return self._do_match(self._peek(), expected_type)

    def _do_match(self, token, expected_type):
        if token is None:
            return False
        if isinstance(expected_type, list):
            return any(token['type'] == t for t in expected_type)
        return token['type'] == expected_type

    def _advance_if_match(self, token_type):
        if self._match(token_type):
            self._advance()
        else:
            raise UnexpectedTokenError(
                token=self._current,
                expected_type=token_type,
                expression=self._expression,
            )



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


def register_kinesis_list_streams_pagination_backcompat(event_emitter):
    # The ListStreams previously used the ExclusiveStartStreamName parameter
    # for input tokens to pagination. This operation was then updated to
    # also allow for the typical NextToken input and output parameters. The
    # pagination model was also updated to use the NextToken field instead of
    # the ExclusiveStartStreamName field for input tokens. However, the
    # ExclusiveStartStreamName is still a valid parameter to control pagination
    # of this operation and is incompatible with the NextToken parameter. So,
    # the CLI needs to continue to treat the ExclusiveStartStreamName as if it
    # is a raw input token parameter to the API by disabling auto-pagination if
    # provided. Otherwise, if it was treated as a normal API parameter, errors
    # would be thrown when paginating across multiple pages since the parameter
    # is incompatible with the NextToken parameter.
    event_emitter.register(
        'building-argument-table.kinesis.list-streams',
        undocument_exclusive_start_stream_name,
    )
    event_emitter.register(
        'operation-args-parsed.kinesis.list-streams',
        disable_pagination_when_exclusive_start_stream_name_provided,
    )


def undocument_exclusive_start_stream_name(argument_table, **kwargs):
    argument_table['exclusive-start-stream-name']._UNDOCUMENTED = True


def disable_pagination_when_exclusive_start_stream_name_provided(
    parsed_args, parsed_globals, **kwargs
):
    if parsed_args.exclusive_start_stream_name is not None:
        parsed_globals.paginate = False



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

from awscli.customizations.emr import configutils, exceptions

LOG = logging.getLogger(__name__)

SUPPORTED_CONFIG_LIST = [
    {'name': 'service_role'},
    {'name': 'log_uri'},
    {
        'name': 'instance_profile',
        'arg_name': 'ec2_attributes',
        'arg_value_key': 'InstanceProfile',
    },
    {
        'name': 'key_name',
        'arg_name': 'ec2_attributes',
        'arg_value_key': 'KeyName',
    },
    {'name': 'enable_debugging', 'type': 'boolean'},
    {'name': 'key_pair_file'},
]

TYPES = ['string', 'boolean']


def get_applicable_configurations(command):
    supported_configurations = _create_supported_configurations()
    return [x for x in supported_configurations if x.is_applicable(command)]


def _create_supported_configuration(config):
    config_type = config['type'] if 'type' in config else 'string'

    if config_type == 'string':
        config_arg_name = (
            config['arg_name'] if 'arg_name' in config else config['name']
        )
        config_arg_value_key = (
            config['arg_value_key'] if 'arg_value_key' in config else None
        )
        configuration = StringConfiguration(
            config['name'], config_arg_name, config_arg_value_key
        )
    elif config_type == 'boolean':
        configuration = BooleanConfiguration(config['name'])

    return configuration


def _create_supported_configurations():
    return [
        _create_supported_configuration(config)
        for config in SUPPORTED_CONFIG_LIST
    ]


class Configuration:
    def __init__(self, name, arg_name):
        self.name = name
        self.arg_name = arg_name

    def is_applicable(self, command):
        raise NotImplementedError("is_applicable")

    def is_present(self, parsed_args):
        raise NotImplementedError("is_present")

    def add(self, command, parsed_args, value):
        raise NotImplementedError("add")

    def _check_arg(self, parsed_args, arg_name):
        return getattr(parsed_args, arg_name, None)


class StringConfiguration(Configuration):
    def __init__(self, name, arg_name, arg_value_key=None):
        super(StringConfiguration, self).__init__(name, arg_name)
        self.arg_value_key = arg_value_key

    def is_applicable(self, command):
        return command.supports_arg(self.arg_name.replace('_', '-'))

    def is_present(self, parsed_args):
        if not self.arg_value_key:
            return self._check_arg(parsed_args, self.arg_name)
        else:
            return self._check_arg(
                parsed_args, self.arg_name
            ) and self.arg_value_key in getattr(parsed_args, self.arg_name)

    def add(self, command, parsed_args, value):
        if not self.arg_value_key:
            setattr(parsed_args, self.arg_name, value)
        else:
            if not self._check_arg(parsed_args, self.arg_name):
                setattr(parsed_args, self.arg_name, {})
            getattr(parsed_args, self.arg_name)[self.arg_value_key] = value


class BooleanConfiguration(Configuration):
    def __init__(self, name):
        super(BooleanConfiguration, self).__init__(name, name)
        self.no_version_arg_name = "no_" + name

    def is_applicable(self, command):
        return command.supports_arg(
            self.arg_name.replace('_', '-')
        ) and command.supports_arg(self.no_version_arg_name.replace('_', '-'))

    def is_present(self, parsed_args):
        return self._check_arg(parsed_args, self.arg_name) or self._check_arg(
            parsed_args, self.no_version_arg_name
        )

    def add(self, command, parsed_args, value):
        if value.lower() == 'true':
            setattr(parsed_args, self.arg_name, True)
            setattr(parsed_args, self.no_version_arg_name, False)
        elif value.lower() == 'false':
            setattr(parsed_args, self.arg_name, False)
            setattr(parsed_args, self.no_version_arg_name, True)
        else:
            raise exceptions.InvalidBooleanConfigError(
                config_value=value,
                config_key=self.arg_name,
                profile_var_name=configutils.get_current_profile_var_name(
                    command._session
                ),
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

import prompt_toolkit
from prompt_toolkit.completion import Completer, Completion

from awscli.customizations.wizard.ui import selectmenu


class Prompter:
    def prompt(self, display_text, choices=None):
        raise NotImplementedError('prompt')


class UIPrompter(Prompter):
    def prompt(self, display_text, choices=None):
        if not choices:
            return prompt_toolkit.prompt('%s: ' % display_text)
        else:
            prompt_toolkit.print_formatted_text(display_text)
            if isinstance(choices[0], str):
                return selectmenu.select_menu(choices)
            else:
                response = selectmenu.select_menu(
                    choices, display_format=self._display_text
                )
                result = response['actual_value']
                return result

    def _display_text(self, obj):
        return obj['display']


class FileCompleter(Completer):
    def get_completions(self, document, complete_event):
        path = document.text
        dirname, partial = os.path.split(path)
        full_dirname = os.path.expanduser(dirname)
        try:
            children = os.listdir(full_dirname)
            for child in sorted(children):
                if child.startswith(partial):
                    result = os.path.join(dirname, child)
                    yield Completion(result, start_position=-len(result))
        except OSError:
            return


class UIFilePrompter:
    def __init__(self, completer):
        self._completer = completer

    def prompt(self, display_text):
        return prompt_toolkit.prompt(
            '%s: ' % display_text, completer=self._completer
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from awscli.customizations.emr import (
    argumentschema,
    emrutils,
    helptext,
    instancegroupsutils,
)
from awscli.customizations.emr.command import Command


class AddInstanceGroups(Command):
    NAME = 'add-instance-groups'
    DESCRIPTION = 'Adds an instance group to a running cluster.'
    ARG_TABLE = [
        {
            'name': 'cluster-id',
            'required': True,
            'help_text': helptext.CLUSTER_ID,
        },
        {
            'name': 'instance-groups',
            'required': True,
            'help_text': helptext.INSTANCE_GROUPS,
            'schema': argumentschema.INSTANCE_GROUPS_SCHEMA,
        },
    ]

    def _run_main_command(self, parsed_args, parsed_globals):
        parameters = {'JobFlowId': parsed_args.cluster_id}
        parameters['InstanceGroups'] = (
            instancegroupsutils.build_instance_groups(
                parsed_args.instance_groups
            )
        )

        add_instance_groups_response = emrutils.call(
            self._session,
            'add_instance_groups',
            parameters,
            self.region,
            parsed_globals.endpoint_url,
            parsed_globals.verify_ssl,
        )

        constructed_result = self._construct_result(
            add_instance_groups_response
        )

        emrutils.display_response(
            self._session,
            'add_instance_groups',
            constructed_result,
            parsed_globals,
        )
        return 0

    def _construct_result(self, add_instance_groups_result):
        jobFlowId = None
        instanceGroupIds = None
        clusterArn = None
        if add_instance_groups_result is not None:
            jobFlowId = add_instance_groups_result.get('JobFlowId')
            instanceGroupIds = add_instance_groups_result.get(
                'InstanceGroupIds'
            )
            clusterArn = add_instance_groups_result.get('ClusterArn')

        if jobFlowId is not None and instanceGroupIds is not None:
            return {
                'ClusterId': jobFlowId,
                'InstanceGroupIds': instanceGroupIds,
                'ClusterArn': clusterArn,
            }
        else:
            return {}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

from awscli.customizations.commands import BasicCommand
from awscli.customizations.utils import uni_print


class ListProfilesCommand(BasicCommand):
    NAME = 'list-profiles'
    DESCRIPTION = 'List the profiles available to the AWS CLI.'
    EXAMPLES = 'aws configure list-profiles\n\n'

    def __init__(self, session, out_stream=None):
        super(ListProfilesCommand, self).__init__(session)

        if out_stream is None:
            out_stream = sys.stdout
        self._out_stream = out_stream

    def _run_main(self, parsed_args, parsed_globals):
        for profile in self._session.available_profiles:
            uni_print('%s\n' % profile, out_file=self._out_stream)
        return 0



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import errno
import os
import shutil
import sys

from awscli.customizations.codedeploy.utils import (
    validate_instance,
    validate_region,
    validate_s3_location,
)
from awscli.customizations.commands import BasicCommand


class Install(BasicCommand):
    NAME = 'install'

    DESCRIPTION = (
        'Configures and installs the AWS CodeDeploy Agent on the on-premises '
        'instance.'
    )

    ARG_TABLE = [
        {
            'name': 'config-file',
            'synopsis': '--config-file <path>',
            'required': True,
            'help_text': (
                'Required. The path to the on-premises instance configuration '
                'file.'
            ),
        },
        {
            'name': 'override-config',
            'action': 'store_true',
            'default': False,
            'help_text': (
                'Optional. Overrides the on-premises instance configuration '
                'file.'
            ),
        },
        {
            'name': 'agent-installer',
            'synopsis': '--agent-installer <s3-location>',
            'required': False,
            'help_text': (
                'Optional. The AWS CodeDeploy Agent installer file.'
            ),
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        params = parsed_args
        params.session = self._session
        validate_region(params, parsed_globals)
        validate_instance(params)
        params.system.validate_administrator()
        self._validate_override_config(params)
        self._validate_agent_installer(params)

        try:
            self._create_config(params)
            self._install_agent(params)
        except Exception as e:
            sys.stdout.flush()
            sys.stderr.write(
                'ERROR\n'
                f'{e}\n'
                'Install the AWS CodeDeploy Agent on the on-premises instance '
                'by following the instructions in "Configure Existing '
                'On-Premises Instances by Using AWS CodeDeploy" in the AWS '
                'CodeDeploy User Guide.\n'
            )
            return 255
        return 0

    def _validate_override_config(self, params):
        if (
            os.path.isfile(params.system.CONFIG_PATH)
            and not params.override_config
        ):
            raise RuntimeError(
                'The on-premises instance configuration file already exists. '
                'Specify --override-config to update the existing on-premises '
                'instance configuration file.'
            )

    def _validate_agent_installer(self, params):
        validate_s3_location(params, 'agent_installer')
        if 'bucket' not in params:
            params.bucket = f'aws-codedeploy-{params.region}'
        if 'key' not in params:
            params.key = f'latest/{params.system.INSTALLER}'
            params.installer = params.system.INSTALLER
        else:
            start = params.key.rfind('/') + 1
            params.installer = params.key[start:]

    def _create_config(self, params):
        sys.stdout.write(
            'Creating the on-premises instance configuration file... '
        )
        try:
            os.makedirs(params.system.CONFIG_DIR)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise e
        if params.config_file != params.system.CONFIG_PATH:
            shutil.copyfile(params.config_file, params.system.CONFIG_PATH)
        sys.stdout.write('DONE\n')

    def _install_agent(self, params):
        sys.stdout.write('Installing the AWS CodeDeploy Agent... ')
        params.system.install(params)
        sys.stdout.write('DONE\n')



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.paginate import (
    MAX_ITEMS_HELP,
    PAGE_SIZE_HELP,
    STARTING_TOKEN_HELP,
)

TABLE_NAME = {
    'name': 'table_name',
    'positional_arg': True,
    'synopsis': '<table-name>',
    'help_text': '<p>The name of your DynamoDB table.</p>',
}

INDEX_NAME = {
    'name': 'index-name',
    'help_text': (
        '<p>The name of a secondary index to scan. This index can be any '
        'local secondary index or global secondary index.</p>'
    ),
}

SELECT = {
    'name': 'attributes',
    'choices': ['ALL', 'ALL_PROJECTED', 'COUNT'],
    'help_text': (
        '<p>The attributes to be returned in the result. You can retrieve all '
        'item attributes, specific item attributes, the count of matching '
        'items, or in the case of an index, some or all of the attributes '
        'projected into the index.</p>'
        '<ul>'
        '<li> ``ALL`` - Returns all of the item attributes from the specified '
        'table or index. If you query a local secondary index, then for each '
        'matching item in the index DynamoDB will fetch the entire item from '
        'the parent table. If the index is configured to project all item '
        'attributes, then all of the data can be obtained from the local '
        'secondary index, and no fetching is required.</li>'
        '<li> ``ALL_PROJECTED`` - Allowed only when querying an index. '
        'Retrieves all attributes that have been projected into the index. '
        'If the index is configured to project all attributes, this return '
        'value is equivalent to specifying ``ALL``.</li>'
        '<li> ``COUNT`` - Returns the number of matching items, rather than '
        'the matching items themselves.</li>'
        '</ul>'
    ),
}

PROJECTION_EXPRESSION = {
    'name': 'projection',
    'nargs': '+',
    'help_text': (
        '<p>A string that identifies one or more attributes to retrieve from '
        'the specified table or index. These attributes can include scalars, '
        'sets, or elements of a JSON document. The attributes in the '
        'expression must be separated by commas. If any of the requested '
        'attributes are not found, they will not appear in the result.</p>'
        '<p>For more information, see '
        '<a href="https://docs.aws.amazon.com/amazondynamodb/latest/'
        'developerguide/Expressions.ProjectionExpressions.html">'
        'Accessing Item Attributes</a> in the <i>Amazon DynamoDB Developer '
        'Guide</i>.'
        '<p>For CLI specific syntax see '
        '<a href="https://docs.aws.amazon.com/cli/latest/topic/'
        'ddb-expressions.html">aws help ddb-expressions</a></p>'
    ),
}

FILTER_EXPRESSION = {
    'name': 'filter',
    'nargs': '+',
    'help_text': (
        '<p>A string that contains conditions that DynamoDB applies after the '
        'operation, but before the data is returned to you. Items that do '
        'not satisfy the ``--filter`` criteria are not returned.</p>'
        '<note><p>A ``--filter`` is applied after the items have already been '
        'read; the process of filtering does not consume any additional read '
        'capacity units.</p></note>'
        '<p>For more information, see '
        '<a href="http://docs.aws.amazon.com/amazondynamodb/latest/'
        'developerguide/QueryAndScan.html#FilteringResults">Filter '
        'Expressions</a> in the <i>Amazon DynamoDB Developer Guide</i>.</p>'
        '<p>For CLI specific syntax see '
        '<a href="https://docs.aws.amazon.com/cli/latest/topic/'
        'ddb-expressions.html">aws help ddb-expressions</a></p>'
    ),
}

CONDITION_EXPRESSION = {
    'name': 'condition',
    'nargs': '+',
    'help_text': (
        '<p>A condition that must be satisfied in order for a conditional '
        '<code>put</code> operation to succeed.</p>'
        '<p>For more information, see '
        '<a href="https://docs.aws.amazon.com/amazondynamodb/latest/'
        'developerguide/Expressions.OperatorsAndFunctions.html">Comparison '
        'Operator and Function Reference</a> in the <i>Amazon DynamoDB '
        'Developer Guide</i></p>'
        '<p>For CLI specific syntax see '
        '<a href="https://docs.aws.amazon.com/cli/latest/topic/'
        'ddb-expressions.html">aws help ddb-expressions</a></p>'
    ),
}

KEY_CONDITION_EXPRESSION = {
    'name': 'key-condition',
    'nargs': '+',
    'help_text': (
        '<p>The condition that specifies the key value(s) for items to be '
        'retrieved. Must perform an equality test on a single partition key '
        'value.</p> <p>The condition can optionally perform one of several '
        'comparison tests on a single sort key value. This allows '
        '<code>select</code> to retrieve one item with a given partition key '
        'value and sort key value, or several items that have the same '
        'partition key value but different sort key values.</p>'
        '<p>The partition key equality test must be specified in the '
        'following format:</p>'
        '<p><code>partitionKeyName = :partitionkeyval</code></p>'
        '<p>If you also want to provide a condition for the sort key, it must '
        'be combined using <code>AND</code> with the condition for the sort '
        'key.</p>'
        '<p>Valid comparisons for the sort key condition are as follows:</p>'
        '<ul>'
        '<li><p><code>sortKeyName = :sortkeyval</code> - true '
        'if the sort key value is equal to <code>:sortkeyval</code>.</p></li>'
        '<li><p><code>sortKeyName &lt; :sortkeyval</code> - true if the sort '
        'key value is less than <code>:sortkeyval</code>.</p></li>'
        '<li><p><code>sortKeyName &lt;= :sortkeyval</code> - true if the '
        'sort key value is less than or equal to <code>:sortkeyval</code>.'
        '</p></li>'
        '<li><p><code>sortKeyName &gt; :sortkeyval</code> - true if the sort '
        'key value is greater than <code>:sortkeyval</code>.</p></li>'
        '<li><p><code>sortKeyName &gt;= :sortkeyval</code> - true if the '
        'sort key value is greater than or equal to <code>:sortkeyval</code>.'
        '</p></li>'
        '<li><p><code>sortKeyName BETWEEN :sortkeyval1 AND '
        ':sortkeyval2</code> - true if the sort key value is greater than or '
        'equal to <code>:sortkeyval1</code>, and less than or equal to '
        '<code>:sortkeyval2</code>.</p></li>'
        '<li><p><code>begins_with(sortKeyName, :sortkeyval)</code> - true '
        'if the sort key value begins with a particular operand. '
        '(You cannot use this function with a sort key that is of type '
        'Number.) Note that the function name <code>begins_with</code> is '
        'case-sensitive.</p></li>'
        '</ul>'
        '<p>For CLI specific syntax see '
        '<a href="https://docs.aws.amazon.com/cli/latest/topic/'
        'ddb-expressions.html">aws help ddb-expressions</a></p>'
    ),
}

ITEMS = {
    'name': 'items',
    'positional_arg': True,
    'synopsis': '<items>',
    'help_text': (
        '<p>One or more items to put into the table, in YAML format.</p>'
    ),
}

CONSISTENT_READ = {
    'name': 'consistent-read',
    'action': 'store_true',
    'default': True,
    'group_name': 'consistent_read',
    'dest': 'consistent_read',
    'help_text': (
        '<p>Determines the read consistency model: If set to '
        '<code>--consistent-read</code>, then the operation uses strongly '
        'consistent reads; otherwise, the operation uses eventually '
        'consistent reads. Strongly consistent reads are not supported on '
        'global secondary indexes. If you query a global secondary index '
        'with <code>--consistent-read</code>, you will receive a '
        '<code>ValidationException</code>.</p>'
    ),
}

NO_CONSISTENT_READ = {
    'name': 'no-consistent-read',
    'action': 'store_false',
    'default': True,
    'group_name': 'consistent_read',
    'dest': 'consistent_read',
}

RETURN_CONSUMED_CAPACITY = {
    'name': 'return-consumed-capacity',
    'action': 'store_true',
    'default': False,
    'group_name': 'return_consumed_capacity',
    'dest': 'return_consumed_capacity',
    'help_text': (
        '<p>Will include the aggregate <code>ConsumedCapacity</code> for the '
        'operation. If <code>--index-name</code> is also specified, '
        'then the <code>ConsumedCapacity</code> for each table and secondary '
        'index that was accessed will be returned.</p>'
    ),
}

NO_RETURN_CONSUMED_CAPACITY = {
    'name': 'no-return-consumed-capacity',
    'action': 'store_false',
    'default': False,
    'group_name': 'return_consumed_capacity',
    'dest': 'return_consumed_capacity',
}

MAX_ITEMS = {
    'name': 'max-items',
    'cli_type_name': 'integer',
    'help_text': MAX_ITEMS_HELP,
}

PAGE_SIZE = {
    'name': 'page-size',
    'cli_type_name': 'integer',
    'help_text': PAGE_SIZE_HELP,
}

STARTING_TOKEN = {
    'name': 'starting-token',
    'help_text': STARTING_TOKEN_HELP,
}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import binascii
import re
import string
from base64 import b64decode
from decimal import Decimal

from .exceptions import EmptyExpressionError, LexerError
from .types import Binary

VALID_BASE64 = re.compile(r'[A-Za-z0-9+/=]+')


class Lexer:
    START_IDENTIFIER = set(string.ascii_letters + '_')
    VALID_IDENTIFIER = set(string.ascii_letters + string.digits + '_')
    WHITESPACE = set(' \t\n\r')
    SIMPLE_TOKENS = {
        '.': 'dot',
        ',': 'comma',
        ':': 'colon',
        '(': 'lparen',
        ')': 'rparen',
        '{': 'lbrace',
        '}': 'rbrace',
        '[': 'lbracket',
        ']': 'rbracket',
        '=': 'eq',
    }
    DIGITS = set(string.digits)
    INT_CHARS = set(string.digits + '-')
    OPERATION_TOKENS = {
        'and',
        'between',
        'in',
        'or',
        'not',
    }

    def tokenize(self, expression):
        self._init_expression(expression)
        while self._current is not None:
            if self._current in self.SIMPLE_TOKENS:
                yield {
                    'type': self.SIMPLE_TOKENS[self._current],
                    'value': self._current,
                    'start': self._position,
                    'end': self._position + 1,
                }
                self._next()
            elif self._current in self.START_IDENTIFIER:
                yield self._consume_unquoted_identifier()
            elif self._current == "'":
                yield self._consume_quoted_identifier()
            elif self._current == '"':
                yield self._consume_string_literal()
            elif self._current in self.WHITESPACE:
                self._next()
            elif self._current in self.INT_CHARS:
                yield self._consume_number()
            elif self._current in ['<', '>']:
                yield self._consume_comparator()
            else:
                raise LexerError(
                    message='Unrecognized character %s' % self._current,
                    expression=self._expression,
                    position=self._position,
                )
        yield {
            'type': 'eof',
            'value': '',
            'start': self._length,
            'end': self._length,
        }

    def _consume_unquoted_identifier(self):
        start = self._position
        buff = self._current

        if self._current in {'b', "B"}:
            if self._next() == '"':
                return self._consume_base64_string()
            elif self._current in self.VALID_IDENTIFIER:
                buff += self._current
            else:
                return {
                    'type': 'unquoted_identifier',
                    'value': buff,
                    'start': start,
                    'end': start + 1,
                }

        while self._next() in self.VALID_IDENTIFIER:
            buff += self._current

        lower = buff.lower()
        if lower in self.OPERATION_TOKENS:
            return {
                'type': lower,
                'value': buff,
                'start': start,
                'end': start + len(buff),
            }

        if lower == 'true':
            return {
                'type': 'literal',
                'value': True,
                'start': start,
                'end': start + len(buff),
            }

        if lower == 'false':
            return {
                'type': 'literal',
                'value': False,
                'start': start,
                'end': start + len(buff),
            }

        if lower == 'null':
            return {
                'type': 'literal',
                'value': None,
                'start': start,
                'end': start + len(buff),
            }
        return {
            'type': 'unquoted_identifier',
            'value': buff,
            'start': start,
            'end': start + len(buff),
        }

    def _consume_quoted_identifier(self):
        start = self._position
        lexeme = self._consume_until("'").replace("\\'", "'")
        token_len = self._position - start
        return {
            'type': 'identifier',
            'value': lexeme,
            'start': start,
            'end': token_len,
        }

    def _consume_string_literal(self):
        start = self._position
        lexeme = self._consume_until('"').replace('\\"', '"')
        token_len = self._position - start
        return {
            'type': 'literal',
            'value': lexeme,
            'start': start,
            'end': token_len,
        }

    def _consume_base64_string(self):
        start = self._position - 1
        raw_string = self._consume_string_literal()

        # Python will simply ignore invalid characters, so we have to
        # validate manually.
        if raw_string['value'] and not VALID_BASE64.match(raw_string['value']):
            message = 'Invalid base64 string b"%s"'
            raise LexerError(
                message=message % raw_string['value'],
                expression=self._expression,
                position=start,
            )

        try:
            decoded = b64decode(raw_string['value'])
        except (TypeError, binascii.Error) as e:
            message = 'Invalid base64 string b"%s": %s'
            message = message % (raw_string['value'], str(e))
            raise LexerError(
                message=message,
                expression=self._expression,
                position=start,
            )

        raw_string['value'] = Binary(decoded)
        raw_string['start'] = start
        return raw_string

    def _consume_number(self):
        start = self._position
        buff = self._consume_int()

        if self._current == '.':
            buff += self._current
            if self._next() not in self.DIGITS:
                raise LexerError(
                    message='Invalid fractional character %s' % self._current,
                    position=self._position,
                    expression=self._expression,
                )
            buff += self._consume_int()

        if self._current and self._current.lower() == 'e':
            buff += self._current
            if self._next() not in self.INT_CHARS and self._current != '+':
                raise LexerError(
                    message='Invalid exponential character %s' % self._current,
                    position=self._position,
                    expression=self._expression,
                )
            buff += self._consume_int()

        return {
            'type': 'literal',
            'value': Decimal(buff),
            'start': start,
            'end': start + len(buff),
        }

    def _consume_int(self):
        buff = self._current
        is_positive = self._current != '-'
        while self._next() in self.DIGITS:
            buff += self._current
        if not is_positive and len(buff) < 2:
            raise LexerError(
                message='Unknown token %s' % buff,
                position=self._position,
                expression=self._expression,
            )
        return buff

    def _consume_comparator(self):
        if self._current == '<':
            if self._next() == '>':
                self._next()
                return {
                    'type': 'ne',
                    'value': '<>',
                    'start': self._position - 2,
                    'end': self._position,
                }
            elif self._current == '=':
                self._next()
                return {
                    'type': 'lte',
                    'value': '<=',
                    'start': self._position - 2,
                    'end': self._position,
                }
            else:
                return {
                    'type': 'lt',
                    'value': '<',
                    'start': self._position - 1,
                    'end': self._position,
                }
        elif self._current == '>':
            if self._next() == '=':
                self._next()
                return {
                    'type': 'gte',
                    'value': '>=',
                    'start': self._position - 2,
                    'end': self._position,
                }
            else:
                return {
                    'type': 'gt',
                    'value': '>',
                    'start': self._position - 1,
                    'end': self._position,
                }

    def _init_expression(self, expression):
        if not expression:
            raise EmptyExpressionError()
        self._position = 0
        self._expression = expression
        self._chars = list(expression)
        self._current = self._chars[0]
        self._length = len(expression)

    def _next(self):
        if self._position == self._length - 1:
            self._current = None
        else:
            self._position += 1
            self._current = self._chars[self._position]
        return self._current

    def _consume_until(self, delimiter):
        # Consume until the delimiter is reached,
        # allowing for the delimiter to be escaped with "\".
        start = self._position
        buff = ''
        self._next()
        while self._current != delimiter:
            if self._current == '\\':
                buff += '\\'
                self._next()
            if self._current is None:
                # We're at the EOF.
                raise LexerError(
                    message="Unclosed %s delimiter" % delimiter,
                    position=start,
                    expression=self._expression,
                )
            buff += self._current
            self._next()
        # Skip the closing delimiter.
        self._next()
        return buff



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import sys
from base64 import b64decode
from decimal import Decimal

from ruamel.yaml import YAML

import awscli.customizations.dynamodb.params as parameters
from awscli.customizations.commands import BasicCommand, CustomArgument
from awscli.customizations.dynamodb.extractor import AttributeExtractor
from awscli.customizations.dynamodb.formatter import DynamoYAMLDumper
from awscli.customizations.dynamodb.transform import (
    ParameterTransformer,
    TypeDeserializer,
    TypeSerializer,
)
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.paginate import ensure_paging_params_not_set
from awscli.formatter import YAMLFormatter
from awscli.utils import OutputStreamFactory

from .types import Binary

LOGGER = logging.getLogger(__name__)


class DDBCommand(BasicCommand):
    def _run_main(self, parsed_args, parsed_globals):
        factory = self._session.get_component('response_parser_factory')
        factory.set_parser_defaults(blob_parser=None)
        self._client = self._session.create_client(
            'dynamodb',
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        self._transformer = ParameterTransformer()
        self._serializer = TypeSerializer()
        self._deserializer = TypeDeserializer()
        self._extractor = AttributeExtractor()
        self._output_stream_factory = OutputStreamFactory(self._session)

    def _serialize(self, operation_name, data):
        service_model = self._client.meta.service_model
        operation_model = service_model.operation_model(
            self._client.meta.method_to_api_mapping.get(operation_name)
        )
        self._transformer.transform(
            data,
            operation_model.input_shape,
            self._serializer.serialize,
            'AttributeValue',
        )

    def _deserialize(self, operation_name, data):
        service_model = self._client.meta.service_model
        operation_model = service_model.operation_model(
            self._client.meta.method_to_api_mapping.get(operation_name)
        )
        self._transformer.transform(
            data,
            operation_model.output_shape,
            self._deserializer.deserialize,
            'AttributeValue',
        )

    def _make_api_call(
        self, operation_name, client_args, should_paginate=True
    ):
        self._serialize(operation_name, client_args)

        if self._client.can_paginate(operation_name) and should_paginate:
            paginator = self._client.get_paginator(operation_name)
            response = paginator.paginate(**client_args).build_full_result()
        else:
            response = getattr(self._client, operation_name)(**client_args)
        if (
            'ConsumedCapacity' in response
            and response['ConsumedCapacity'] is None
        ):
            del response['ConsumedCapacity']
        self._deserialize(operation_name, response)
        return response

    def _dump_yaml(self, operation_name, data, parsed_globals):
        # TODO: In the future, we should support yaml-stream. However, it
        #  would require a larger refactoring. Right now we always build
        #  the full result when paginating prior to sending it to the
        #  formatter. We need to instead pass the page iterator and
        #  deserialize in the formatter. We cannot necessarily just
        #  convert these to client handlers because the DDB types we
        #  introduce do not play nicely with the pagination interfaces.
        #  For example, botocore cannot serialize our Binary types into
        #  a resume token when --max-items gets set.
        formatter = YAMLFormatter(parsed_globals, DynamoYAMLDumper())
        with self._output_stream_factory.get_output_stream() as stream:
            formatter(operation_name, data, stream)

    def _add_expression_args(
        self, expression_name, expression, args, substitution_count=0
    ):
        result = self._extractor.extract(
            ' '.join(expression), substitution_count
        )
        args[expression_name] = result['expression']

        if result['identifiers']:
            if 'ExpressionAttributeNames' not in args:
                args['ExpressionAttributeNames'] = {}
            args['ExpressionAttributeNames'].update(result['identifiers'])

        if result['values']:
            if 'ExpressionAttributeValues' not in args:
                args['ExpressionAttributeValues'] = {}
            args['ExpressionAttributeValues'].update(result['values'])

        return result['substitution_count']


class PaginatedDDBCommand(DDBCommand):
    PAGING_ARGS = [
        parameters.STARTING_TOKEN,
        parameters.MAX_ITEMS,
        parameters.PAGE_SIZE,
    ]

    def _build_arg_table(self):
        arg_table = super(PaginatedDDBCommand, self)._build_arg_table()
        for arg_data in self.PAGING_ARGS:
            paging_arg = CustomArgument(**arg_data)
            arg_table[arg_data['name']] = paging_arg
        return arg_table

    def _get_client_args(self, parsed_args):
        pagination_config = {}
        if parsed_args.starting_token:
            pagination_config['StartingToken'] = parsed_args.starting_token
        if parsed_args.max_items:
            pagination_config['MaxItems'] = parsed_args.max_items
        if parsed_args.page_size:
            pagination_config['PageSize'] = parsed_args.page_size

        args = {}
        if pagination_config:
            args['PaginationConfig'] = pagination_config
        return args


class SelectCommand(PaginatedDDBCommand):
    NAME = 'select'
    DESCRIPTION = (
        '``select`` searches a table or index.\n\n'
        'Under the hood, this operation will use ``query`` if '
        '``--key-condition`` is specified, or ``scan`` otherwise.\n\n'
        'Only ``yaml`` output is supported for this operation.'
    )
    ARG_TABLE = [
        parameters.TABLE_NAME,
        parameters.INDEX_NAME,
        parameters.PROJECTION_EXPRESSION,
        parameters.FILTER_EXPRESSION,
        parameters.KEY_CONDITION_EXPRESSION,
        parameters.SELECT,
        parameters.CONSISTENT_READ,
        parameters.NO_CONSISTENT_READ,
        parameters.RETURN_CONSUMED_CAPACITY,
        parameters.NO_RETURN_CONSUMED_CAPACITY,
    ]
    _SUPPORTED_OUTPUT_TYPES = ('yaml',)

    def _run_main(self, parsed_args, parsed_globals):
        super(SelectCommand, self)._run_main(parsed_args, parsed_globals)
        self._select(parsed_args, parsed_globals)
        return 0

    def _select(self, parsed_args, parsed_globals):
        output_type = parsed_globals.output
        if (
            output_type is not None
            and output_type not in self._SUPPORTED_OUTPUT_TYPES
        ):
            raise ParamValidationError(
                f'{output_type} output format is not supported for ddb commands'
            )

        if parsed_args.key_condition:
            LOGGER.debug(
                "select command using query because --key-condition was "
                "specified"
            )
            operation = 'query'
        else:
            LOGGER.debug(
                "select command using scan because --key-condition was not "
                "specified"
            )
            operation = 'scan'
        should_paginate = parsed_globals.paginate
        if not should_paginate:
            ensure_paging_params_not_set(parsed_args, {})
        client_args = self._get_client_args(parsed_args)
        response = self._make_api_call(operation, client_args, should_paginate)
        self._dump_yaml(operation, response, parsed_globals)

    def _get_client_args(self, parsed_args):
        client_args = super(SelectCommand, self)._get_client_args(parsed_args)
        client_args.update(
            {
                'TableName': parsed_args.table_name,
                'ConsistentRead': parsed_args.consistent_read,
            }
        )
        substitution_count = 0
        if parsed_args.index_name is not None:
            client_args['IndexName'] = parsed_args.index_name
        if parsed_args.projection is not None:
            substitution_count = self._add_expression_args(
                'ProjectionExpression',
                parsed_args.projection,
                client_args,
                substitution_count,
            )
        if parsed_args.filter is not None:
            substitution_count += self._add_expression_args(
                'FilterExpression',
                parsed_args.filter,
                client_args,
                substitution_count,
            )
        if parsed_args.key_condition is not None:
            self._add_expression_args(
                'KeyConditionExpression',
                parsed_args.key_condition,
                client_args,
                substitution_count,
            )
        if parsed_args.attributes is not None:
            select_map = {
                'ALL': 'ALL_ATTRIBUTES',
                'ALL_PROJECTED': 'ALL_PROJECTED_ATTRIBUTES',
            }
            attrs = parsed_args.attributes
            client_args['Select'] = select_map.get(attrs, attrs)

        if parsed_args.return_consumed_capacity:
            # Only return index info when using an index.
            if parsed_args.index_name is not None:
                client_args['ReturnConsumedCapacity'] = 'INDEXES'
            else:
                client_args['ReturnConsumedCapacity'] = 'TOTAL'
        else:
            client_args['ReturnConsumedCapacity'] = 'NONE'

        return client_args


class PutCommand(DDBCommand):
    NAME = 'put'
    DESCRIPTION = '``put`` puts one or more items into a table.'
    ARG_TABLE = [
        parameters.TABLE_NAME,
        parameters.ITEMS,
        parameters.CONDITION_EXPRESSION,
    ]

    def _run_main(self, parsed_args, parsed_globals):
        super(PutCommand, self)._run_main(parsed_args, parsed_globals)
        self._yaml = YAML(typ='safe')
        self._yaml.constructor.add_constructor(
            'tag:yaml.org,2002:binary', self._load_binary
        )
        self._yaml.constructor.add_constructor(
            'tag:yaml.org,2002:float', self._load_number
        )
        self._put(parsed_args)
        return 0

    def _load_binary(self, loader, node):
        return Binary(b64decode(node.value))

    def _load_number(self, loader, node):
        return Decimal(node.value)

    def _put(self, parsed_args):
        items = self._get_items(parsed_args)

        # batch write does not support condition expressions, so if we use
        # that then we just have to call put_item for each item.
        if len(items) > 1 and parsed_args.condition is None:
            self._batch_write(items, parsed_args)
        else:
            if len(items) > 1:
                raise ParamValidationError(
                    '--condition is not supported for multiple items'
                )
            self._put_item(items, parsed_args)

    def _put_item(self, items, parsed_args):
        client_args = self._get_base_args(parsed_args)
        client_args['TableName'] = parsed_args.table_name
        for item in items:
            client_args['Item'] = item
            self._make_api_call('put_item', client_args)
        return {}

    def _batch_write(self, items, parsed_args):
        batch_size = 25
        client_args = self._get_base_args(parsed_args)

        put_requests = [{'PutRequest': {'Item': i}} for i in items]
        while len(put_requests) > 0:
            batch_items = put_requests[:batch_size]
            client_args['RequestItems'] = {parsed_args.table_name: batch_items}
            result = self._make_api_call('batch_write_item', client_args)

            put_requests = put_requests[batch_size:]

            unprocessed_items = result.get('UnprocessedItems', {})
            unprocessed_items = unprocessed_items.get(
                parsed_args.table_name, None
            )
            if unprocessed_items is not None:
                put_requests.extend(unprocessed_items)

    def _get_items(self, parsed_args):
        if parsed_args.items == '-':
            items = self._yaml.load(sys.stdin)
        else:
            items = self._yaml.load(parsed_args.items)
        if not isinstance(items, list):
            items = [items]
        return items

    def _get_base_args(self, parsed_args):
        client_args = {'ReturnConsumedCapacity': 'NONE'}
        if parsed_args.condition is not None:
            self._add_expression_args(
                'ConditionExpression',
                parsed_args.condition,
                client_args,
            )
        return client_args



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

import awscrt.s3
from botocore.client import Config
from botocore.httpsession import DEFAULT_CA_BUNDLE
from s3transfer.crt import (
    BotocoreCRTCredentialsWrapper,
    BotocoreCRTRequestSerializer,
    CRTTransferManager,
    acquire_crt_s3_process_lock,
    create_s3_crt_client,
)
from s3transfer.manager import TransferManager

from awscli.compat import urlparse
from awscli.customizations.s3 import constants
from awscli.customizations.s3.transferconfig import (
    create_transfer_config_from_runtime_config,
)

LOGGER = logging.getLogger(__name__)


class ClientFactory:
    def __init__(self, session):
        self._session = session

    def create_client(self, params, is_source_client=False):
        create_client_kwargs = {'verify': params['verify_ssl']}
        if params.get('sse') == 'aws:kms':
            create_client_kwargs['config'] = Config(signature_version='s3v4')
        region = params['region']
        endpoint_url = params['endpoint_url']
        if is_source_client and params['source_region']:
            if params['paths_type'] == 's3s3':
                region = params['source_region']
                endpoint_url = None

        create_client_kwargs['region_name'] = region
        create_client_kwargs['endpoint_url'] = endpoint_url
        return self._session.create_client('s3', **create_client_kwargs)


class TransferManagerFactory:
    _MAX_IN_MEMORY_CHUNKS = 6
    _CRT_PROCESS_LOCK_NAME = 'aws-cli'

    def __init__(self, session):
        self._session = session
        self._botocore_client_factory = ClientFactory(self._session)

    def create_transfer_manager(
        self, params, runtime_config, botocore_client=None
    ):
        client_type = self._compute_transfer_client_type(
            params, runtime_config
        )
        if client_type == constants.CRT_TRANSFER_CLIENT:
            return self._create_crt_transfer_manager(params, runtime_config)
        else:
            return self._create_classic_transfer_manager(
                params, runtime_config, botocore_client
            )

    def _compute_transfer_client_type(self, params, runtime_config):
        if params.get('paths_type') == 's3s3':
            return constants.CLASSIC_TRANSFER_CLIENT
        preferred_transfer_client = runtime_config.get(
            'preferred_transfer_client', constants.AUTO_RESOLVE_TRANSFER_CLIENT
        )
        if preferred_transfer_client == constants.AUTO_RESOLVE_TRANSFER_CLIENT:
            return self._resolve_transfer_client_type_for_system()
        return preferred_transfer_client

    def _resolve_transfer_client_type_for_system(self):
        transfer_client_type = constants.CLASSIC_TRANSFER_CLIENT
        is_optimized_for_system = awscrt.s3.is_optimized_for_system()
        LOGGER.debug(
            'S3 CRT client optimized for system: %s', is_optimized_for_system
        )
        if is_optimized_for_system:
            is_running = self._is_crt_client_running_in_other_aws_cli_process()
            LOGGER.debug(
                'S3 CRT client running in different AWS CLI process: %s',
                is_running,
            )
            if not is_running:
                transfer_client_type = constants.CRT_TRANSFER_CLIENT
        LOGGER.debug(
            'Auto resolved s3 transfer client to: %s', transfer_client_type
        )
        return transfer_client_type

    def _is_crt_client_running_in_other_aws_cli_process(self):
        # If None is returned from acquiring the CRT process lock, it
        # means the CRT S3 client is currently being used in a different
        # AWS CLI process.
        return self._acquire_crt_s3_process_lock() is None

    def _acquire_crt_s3_process_lock(self):
        return acquire_crt_s3_process_lock(self._CRT_PROCESS_LOCK_NAME)

    def _create_crt_transfer_manager(self, params, runtime_config):
        self._acquire_crt_s3_process_lock()
        return CRTTransferManager(
            self._create_crt_client(params, runtime_config),
            self._create_crt_request_serializer(params),
        )

    def _create_crt_client(self, params, runtime_config):
        config_file_params = self._session.get_scoped_config().get('s3', {})
        create_crt_client_kwargs = {
            'region': self._resolve_region(params),
            'verify': self._resolve_verify(params),
        }
        endpoint_url = params.get('endpoint_url')
        if endpoint_url and urlparse.urlparse(endpoint_url).scheme == 'http':
            create_crt_client_kwargs['use_ssl'] = False
        target_throughput = runtime_config.get('target_bandwidth', None)
        if target_throughput:
            create_crt_client_kwargs['target_throughput'] = target_throughput
        multipart_chunksize = runtime_config.get('multipart_chunksize', None)
        # User didn't explicitly configure `multipart_chunksize`. Set it to
        # `None` and let CRT dynamically calculate the part size.
        if 'multipart_chunksize' not in config_file_params:
            multipart_chunksize = None
        create_crt_client_kwargs['part_size'] = multipart_chunksize
        if params.get('sign_request', True):
            crt_credentials_provider = self._get_crt_credentials_provider()
            create_crt_client_kwargs['crt_credentials_provider'] = (
                crt_credentials_provider
            )
        fio_options = {}
        if (val := runtime_config.get('should_stream')) is not None:
            fio_options['should_stream'] = val
        if (val := runtime_config.get('disk_throughput')) is not None:
            # Convert bytes to gigabits.
            fio_options['disk_throughput_gbps'] = val * 8 / 1_000_000_000
        if (val := runtime_config.get('direct_io')) is not None:
            fio_options['direct_io'] = val
        create_crt_client_kwargs['fio_options'] = fio_options

        return create_s3_crt_client(**create_crt_client_kwargs)

    def _create_crt_request_serializer(self, params):
        return BotocoreCRTRequestSerializer(
            self._session,
            {
                'region_name': self._resolve_region(params),
                'endpoint_url': params.get('endpoint_url'),
            },
        )

    def _create_classic_transfer_manager(
        self, params, runtime_config, client=None
    ):
        if client is None:
            client = self._botocore_client_factory.create_client(params)
        transfer_config = create_transfer_config_from_runtime_config(
            runtime_config
        )
        transfer_config.max_in_memory_upload_chunks = (
            self._MAX_IN_MEMORY_CHUNKS
        )
        transfer_config.max_in_memory_download_chunks = (
            self._MAX_IN_MEMORY_CHUNKS
        )
        LOGGER.debug(
            "Using a multipart threshold of %s and a part size of %s",
            transfer_config.multipart_threshold,
            transfer_config.multipart_chunksize,
        )
        return TransferManager(client, transfer_config)

    def _get_crt_credentials_provider(self):
        botocore_credentials = self._session.get_credentials()
        wrapper = BotocoreCRTCredentialsWrapper(botocore_credentials)
        return wrapper.to_crt_credentials_provider()

    def _resolve_region(self, params):
        region = params.get('region')
        if region is None:
            region = self._session.get_config_variable('region')
        return region

    def _resolve_verify(self, params):
        verify = params.get('verify_ssl')
        if verify is None:
            verify = self._session.get_config_variable('ca_bundle')
        if verify is None:
            verify = DEFAULT_CA_BUNDLE
        return verify



# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""
Error message formatting module for AWS CLI.

This module provides centralized error formatting functionality to ensure
all error messages follow the standard format: aws: [ERROR]: <error-message>

This standardization improves accessibility for users with visual disabilities
and those using assistive technologies like screen readers.
"""

# Standard error prefix format for all AWS CLI error messages
ERROR_PREFIX = "aws: [ERROR]:"


def format_error_message(message, prog="aws"):
    if not message:
        return f"{prog}: [ERROR]:"

    message = message.strip()

    return f"{prog}: [ERROR]: {message}"


def write_error(stderr, message):
    formatted_message = format_error_message(message)
    stderr.write("\n")
    stderr.write(formatted_message)
    stderr.write("\n")



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import collections
import itertools
import json
import logging
import os
import re

import colorama
from botocore import UNSIGNED
from botocore.config import Config
from botocore.configprovider import ConstantProvider
from botocore.exceptions import ProfileNotFound
from botocore.utils import is_valid_endpoint_url
from prompt_toolkit import prompt as ptk_prompt
from prompt_toolkit.application import get_app
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator

from awscli.customizations.configure import (
    get_section_header,
    profile_to_section,
)
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.sso.utils import (
    LOGIN_ARGS,
    BaseSSOCommand,
    PrintOnlyHandler,
    do_sso_login,
    parse_sso_registration_scopes,
)
from awscli.customizations.utils import uni_print
from awscli.customizations.wizard.ui.selectmenu import select_menu
from awscli.formatter import CLI_OUTPUT_FORMATS

logger = logging.getLogger(__name__)

_CMD_PROMPT_USAGE = (
    'To keep an existing value, hit enter when prompted for the value. When '
    'you are prompted for information, the current value will be displayed in '
    '[brackets]. If the config item has no value, it is displayed as '
    '[None] or omitted entirely.\n\n'
)
_CONFIG_EXTRA_INFO = (
    'Note: The configuration is saved in the shared configuration file. '
    'By default, ``~/.aws/config``. For more information, see the '
    '"Configuring the AWS CLI to use AWS IAM Identity Center" section in the '
    'AWS CLI User Guide:'
    '\n\nhttps://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sso.html'
)


class ValidatorWithDefault(Validator):
    def __init__(self, default=None):
        super(ValidatorWithDefault, self).__init__()
        self._default = default

    def _raise_validation_error(self, document, message):
        index = len(document.text)
        raise ValidationError(index, message)


class StartUrlValidator(ValidatorWithDefault):
    def validate(self, document):
        # If there's a default, allow an empty prompt
        if not document.text and self._default:
            return
        if not is_valid_endpoint_url(document.text):
            self._raise_validation_error(document, 'Not a valid Start URL')


class RequiredInputValidator(ValidatorWithDefault):
    def validate(self, document):
        if document.text or self._default:
            return
        self._raise_validation_error(document, 'A value is required')


class ScopesValidator(ValidatorWithDefault):
    def validate(self, document):
        # If there's a default, allow an empty prompt
        if not document.text and self._default:
            return
        if not self._is_comma_separated_list(document.text):
            self._raise_validation_error(
                document, 'Scope values must be separated by commas'
            )

    def _is_comma_separated_list(self, value):
        scopes = value.split(',')
        for scope in scopes:
            if re.findall(r'\s', scope.strip()):
                return False
        return True


class PTKPrompt:
    _DEFAULT_PROMPT_FORMAT = '{prompt_text} [{current_value}]: '

    def __init__(self, prompter=None):
        if prompter is None:
            prompter = ptk_prompt
        self._prompter = prompter

    def _create_completer(self, completions):
        if completions is None:
            completions = []
        completer_kwargs = {
            'words': completions,
            'pattern': re.compile(r'\S+'),
        }
        if isinstance(completions, dict):
            completer_kwargs['meta_dict'] = completions
            completer_kwargs['words'] = list(completions.keys())
        return WordCompleter(**completer_kwargs)

    def get_value(
        self,
        current_value,
        prompt_text='',
        completions=None,
        validator=None,
        toolbar=None,
        prompt_fmt=None,
    ):
        if prompt_fmt is None:
            prompt_fmt = self._DEFAULT_PROMPT_FORMAT
        prompt_string = prompt_fmt.format(
            prompt_text=prompt_text, current_value=current_value
        )
        prompter_kwargs = {
            'validator': validator,
            'validate_while_typing': False,
            'completer': self._create_completer(completions),
            'complete_while_typing': True,
            'style': self._get_prompt_style(),
        }
        if toolbar:
            prompter_kwargs['bottom_toolbar'] = toolbar
            prompter_kwargs['refresh_interval'] = 0.2
        response = self._prompter(prompt_string, **prompter_kwargs)
        # Strip any extra white space
        response = response.strip()
        if not response:
            # If the user hits enter, we return the current/default value
            response = current_value
        return response

    def _get_prompt_style(self):
        return Style.from_dict(
            {
                'bottom-toolbar': 'noreverse',
            }
        )


def display_account(account):
    """Converts an SSO account response into a display string.

    All fields should be present in the API response but we've seen some
    cases where the account name or email address is not set. Considering
    we only need the account name and email address for display purposes
    we should be defensive just in case they don't come back.
    """
    if 'accountName' not in account and 'emailAddress' not in account:
        account_template = '{accountId}'
    elif 'emailAddress' not in account:
        account_template = '{accountName} ({accountId})'
    elif 'accountName' not in account:
        account_template = '{emailAddress} ({accountId})'
    else:
        account_template = '{accountName}, {emailAddress} ({accountId})'
    return account_template.format(**account)


def get_account_sorting_key(account):
    only_account_id = ('accountName' not in account and 'emailAddress' not in account)
    for key in ('accountName', 'emailAddress', 'accountId'):
        value = account.get(key, None)
        if value is not None:
            return (only_account_id, value.lower())
    return (only_account_id, None)


class SSOSessionConfigurationPrompter:
    _DEFAULT_SSO_SCOPE = 'sso:account:access'
    _KNOWN_SSO_SCOPES = {
        'sso:account:access': (
            'Grants access to AWS IAM Identity Center accounts and permission '
            'sets'
        )
    }

    def __init__(self, botocore_session, prompter):
        self._botocore_session = botocore_session
        self._prompter = prompter
        self._sso_sessions = self._botocore_session.full_config.get(
            'sso_sessions', {}
        )
        self._sso_session = None
        self.sso_session_config = {}

    @property
    def sso_session(self):
        return self._sso_session

    @sso_session.setter
    def sso_session(self, value):
        self._sso_session = value
        self.sso_session_config = self._sso_sessions.get(
            self._sso_session, {}
        ).copy()

    def prompt_for_sso_session(self, required=True):
        prompt_text = 'SSO session name'
        prompt_fmt = None
        validator_cls = None
        if required:
            validator_cls = RequiredInputValidator
        if not self.sso_session:
            prompt_fmt = f'{prompt_text}: '
            if not required:
                prompt_fmt = f'{prompt_text} (Recommended): '
        sso_session = self._prompt_for(
            'sso_session',
            prompt_text,
            completions=sorted(self._sso_sessions),
            toolbar=self._get_sso_session_toolbar,
            validator_cls=validator_cls,
            prompt_fmt=prompt_fmt,
            current_value=self.sso_session,
        )
        self.sso_session = sso_session
        return sso_session

    def prompt_for_sso_start_url(self):
        return self._prompt_for(
            'sso_start_url',
            'SSO start URL',
            completions=self._get_potential_start_urls(),
            validator_cls=StartUrlValidator,
        )

    def prompt_for_sso_region(self):
        return self._prompt_for(
            'sso_region',
            'SSO region',
            completions=self._get_potential_sso_regions(),
            validator_cls=RequiredInputValidator,
        )

    def prompt_for_sso_registration_scopes(self):
        if 'sso_registration_scopes' not in self.sso_session_config:
            self.sso_session_config['sso_registration_scopes'] = (
                self._DEFAULT_SSO_SCOPE
            )
        raw_scopes = self._prompt_for(
            'sso_registration_scopes',
            'SSO registration scopes',
            completions=self._get_potential_sso_registrations_scopes(),
            validator_cls=ScopesValidator,
        )
        return parse_sso_registration_scopes(raw_scopes)

    def _prompt_for(
        self,
        config_name,
        text,
        completions=None,
        validator_cls=None,
        toolbar=None,
        prompt_fmt=None,
        current_value=None,
    ):
        if current_value is None:
            current_value = self.sso_session_config.get(config_name)
        validator = None
        if validator_cls:
            validator = validator_cls(current_value)
        value = self._prompter.get_value(
            current_value,
            text,
            completions=completions,
            validator=validator,
            toolbar=toolbar,
            prompt_fmt=prompt_fmt,
        )
        if value:
            self.sso_session_config[config_name] = value
        return value

    def _get_sso_session_toolbar(self):
        current_input = get_app().current_buffer.document.text
        if current_input in self._sso_sessions:
            selected_sso_config = self._sso_sessions[current_input]
            return FormattedText(
                [
                    ('', self._get_toolbar_border()),
                    ('', '\n'),
                    (
                        'bold',
                        f'Configuration for SSO session: {current_input}\n\n',
                    ),
                    ('', json.dumps(selected_sso_config, indent=2)),
                ]
            )

    def _get_toolbar_border(self):
        horizontal_line_char = '\u2500'
        return horizontal_line_char * get_app().output.get_size().columns

    def _get_potential_start_urls(self):
        profiles = self._botocore_session.full_config.get('profiles', {})
        configs_to_search = itertools.chain(
            profiles.values(), self._sso_sessions.values()
        )
        potential_start_urls = set()
        for config_to_search in configs_to_search:
            if 'sso_start_url' in config_to_search:
                start_url = config_to_search['sso_start_url']
                potential_start_urls.add(start_url)
        return list(potential_start_urls)

    def _get_potential_sso_regions(self):
        return self._botocore_session.get_available_regions('sso-oidc')

    def _get_potential_sso_registrations_scopes(self):
        potential_scopes = self._KNOWN_SSO_SCOPES.copy()
        scopes_to_sessions = self._get_previously_used_scopes_to_sso_sessions()
        for scope, sso_sessions in scopes_to_sessions.items():
            if scope not in potential_scopes:
                potential_scopes[scope] = (
                    f'Used in SSO sessions: {", ".join(sso_sessions)}'
                )
        return potential_scopes

    def _get_previously_used_scopes_to_sso_sessions(self):
        scopes_to_sessions = collections.defaultdict(list)
        for sso_session, sso_session_config in self._sso_sessions.items():
            if 'sso_registration_scopes' in sso_session_config:
                parsed_scopes = parse_sso_registration_scopes(
                    sso_session_config['sso_registration_scopes']
                )
                for parsed_scope in parsed_scopes:
                    scopes_to_sessions[parsed_scope].append(sso_session)
        return scopes_to_sessions


class BaseSSOConfigurationCommand(BaseSSOCommand):
    def __init__(self, session, prompter=None, config_writer=None):
        super(BaseSSOConfigurationCommand, self).__init__(session)
        if prompter is None:
            prompter = PTKPrompt()
        self._prompter = prompter
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer
        self._sso_sessions = self._session.full_config.get('sso_sessions', {})
        self._sso_session_prompter = SSOSessionConfigurationPrompter(
            botocore_session=session,
            prompter=self._prompter,
        )

    def _write_sso_configuration(self):
        self._update_section(
            section_header=get_section_header(
                'sso-session', self._sso_session_prompter.sso_session
            ),
            new_values=self._sso_session_prompter.sso_session_config,
        )

    def _update_section(self, section_header, new_values):
        config_path = self._session.get_config_variable('config_file')
        config_path = os.path.expanduser(config_path)
        new_values['__section__'] = section_header
        self._config_writer.update_config(new_values, config_path)


class ConfigureSSOCommand(BaseSSOConfigurationCommand):
    NAME = 'sso'
    SYNOPSIS = 'aws configure sso [--profile profile-name]'
    DESCRIPTION = (
        'The ``aws configure sso`` command interactively prompts for the '
        'configuration values required to create a profile that sources '
        'temporary AWS credentials from AWS IAM Identity Center.\n\n'
        f'{_CMD_PROMPT_USAGE}'
        'When providing the ``--profile`` parameter the named profile '
        'will be created or updated. When a profile is not explicitly set '
        'the profile name will be prompted for.\n\n'
        f'{_CONFIG_EXTRA_INFO}'
    )
    # TODO: Add CLI parameters to skip prompted values, --start-url, etc.
    ARG_TABLE = LOGIN_ARGS

    def __init__(
        self,
        session,
        prompter=None,
        selector=None,
        config_writer=None,
        sso_token_cache=None,
        sso_login=None,
    ):
        super(ConfigureSSOCommand, self).__init__(
            session, prompter=prompter, config_writer=config_writer
        )
        if selector is None:
            selector = select_menu
        self._selector = selector
        if sso_login is None:
            sso_login = do_sso_login
        self._sso_login = sso_login
        self._sso_token_cache = sso_token_cache

        self._new_profile_config_values = {}
        self._original_profile_name = self._session.profile
        try:
            self._profile_config = self._session.get_scoped_config()
        except ProfileNotFound:
            self._profile_config = {}
        self._set_sso_session_if_configured_in_profile()

    def _set_sso_session_if_configured_in_profile(self):
        if 'sso_session' in self._profile_config:
            self._sso_session_prompter.sso_session = self._profile_config[
                'sso_session'
            ]

    def _handle_single_account(self, accounts):
        sso_account_id = accounts[0]['accountId']
        single_account_msg = 'The only AWS account available to you is: {}\n'
        uni_print(single_account_msg.format(sso_account_id))
        return sso_account_id

    def _handle_multiple_accounts(self, accounts):
        available_accounts_msg = (
            'There are {} AWS accounts available to you.\n'
        )
        uni_print(available_accounts_msg.format(len(accounts)))
        sorted_accounts = sorted(accounts, key=get_account_sorting_key)
        selected_account = self._selector(
            sorted_accounts, display_format=display_account
        )
        sso_account_id = selected_account['accountId']
        return sso_account_id

    def _get_all_accounts(self, sso, sso_token):
        paginator = sso.get_paginator('list_accounts')
        results = paginator.paginate(accessToken=sso_token['accessToken'])
        return results.build_full_result()

    def _prompt_for_account(self, sso, sso_token):
        accounts = self._get_all_accounts(sso, sso_token)['accountList']
        if not accounts:
            raise RuntimeError('No AWS accounts are available to you.')
        if len(accounts) == 1:
            sso_account_id = self._handle_single_account(accounts)
        else:
            sso_account_id = self._handle_multiple_accounts(accounts)
        uni_print(f'Using the account ID {sso_account_id}\n')
        self._new_profile_config_values['sso_account_id'] = sso_account_id
        return sso_account_id

    def _handle_single_role(self, roles):
        sso_role_name = roles[0]['roleName']
        available_roles_msg = 'The only role available to you is: {}\n'
        uni_print(available_roles_msg.format(sso_role_name))
        return sso_role_name

    def _handle_multiple_roles(self, roles):
        available_roles_msg = 'There are {} roles available to you.\n'
        uni_print(available_roles_msg.format(len(roles)))
        sorted_roles = sorted(roles, key=lambda x: x['roleName'].lower())
        role_names = [r['roleName'] for r in sorted_roles]
        sso_role_name = self._selector(role_names)
        return sso_role_name

    def _get_all_roles(self, sso, sso_token, sso_account_id):
        paginator = sso.get_paginator('list_account_roles')
        results = paginator.paginate(
            accountId=sso_account_id, accessToken=sso_token['accessToken']
        )
        return results.build_full_result()

    def _prompt_for_role(self, sso, sso_token, sso_account_id):
        roles = self._get_all_roles(sso, sso_token, sso_account_id)['roleList']
        if not roles:
            error_msg = 'No roles are available for the account {}'
            raise RuntimeError(error_msg.format(sso_account_id))
        if len(roles) == 1:
            sso_role_name = self._handle_single_role(roles)
        else:
            sso_role_name = self._handle_multiple_roles(roles)
        uni_print(f'Using the role name "{sso_role_name}"\n')
        self._new_profile_config_values['sso_role_name'] = sso_role_name
        return sso_role_name

    def _prompt_for_profile(self, sso_account_id=None, sso_role_name=None):
        if self._original_profile_name:
            profile_name = self._original_profile_name
        else:
            text = 'Profile name'
            default_profile = None
            if sso_account_id and sso_role_name:
                default_profile = f'{sso_role_name}-{sso_account_id}'
            validator = RequiredInputValidator(default_profile)
            profile_name = self._prompter.get_value(
                default_profile, text, validator=validator
            )
        return profile_name

    def _prompt_for_cli_default_region(self):
        # TODO: figure out a way to get a list of reasonable client regions
        return self._prompt_for_profile_config(
            'region', 'Default client Region'
        )

    def _prompt_for_cli_output_format(self):
        return self._prompt_for_profile_config(
            'output',
            'CLI default output format (json if not specified)',
            completions=list(CLI_OUTPUT_FORMATS.keys()),
        )

    def _prompt_for_profile_config(self, config_name, text, completions=None):
        current_value = self._profile_config.get(config_name)

        new_value = self._prompter.get_value(
            current_value,
            text,
            completions=completions,
        )
        if new_value:
            self._new_profile_config_values[config_name] = new_value
        return new_value

    def _unset_session_profile(self):
        # The profile provided to the CLI as --profile may not exist.
        # This means we cannot use the session as is to create clients.
        # By overriding the profile provider we ensure that a non-existant
        # profile won't cause us to fail to create clients.
        # No configuration from the profile is needed for the SSO APIs.
        # It might be good to see if we can address this in a better way
        # in botocore.
        config_store = self._session.get_component('config_store')
        config_store.set_config_provider('profile', ConstantProvider(None))

    def _run_main(self, parsed_args, parsed_globals):
        self._unset_session_profile()
        on_pending_authorization = None
        if parsed_args.no_browser:
            on_pending_authorization = PrintOnlyHandler()
        sso_registration_args = self._prompt_for_sso_registration_args()
        sso_token = self._sso_login(
            self._session,
            parsed_globals=parsed_globals,
            token_cache=self._sso_token_cache,
            on_pending_authorization=on_pending_authorization,
            use_device_code=parsed_args.use_device_code,
            **sso_registration_args,
        )

        # Construct an SSO client to explore the accounts / roles
        client_config = Config(
            signature_version=UNSIGNED,
            region_name=sso_registration_args['sso_region'],
        )
        sso = self._session.create_client('sso', config=client_config)

        sso_account_id, sso_role_name = self._prompt_for_sso_account_and_role(
            sso, sso_token
        )
        configured_for_aws_credentials = all((sso_account_id, sso_role_name))

        # General CLI configuration
        self._prompt_for_cli_default_region()
        self._prompt_for_cli_output_format()

        profile_name = self._prompt_for_profile(sso_account_id, sso_role_name)

        self._write_new_config(profile_name)
        self._print_conclusion(configured_for_aws_credentials, profile_name)
        return 0

    def _prompt_for_sso_registration_args(self):
        sso_session = self._sso_session_prompter.prompt_for_sso_session(
            required=False
        )
        if sso_session is None:
            self._warn_configuring_using_legacy_format()
            return self._prompt_for_registration_args_with_legacy_format()
        else:
            self._set_sso_session_in_profile_config(sso_session)
            if sso_session in self._sso_sessions:
                return self._get_sso_registration_args_from_sso_config(
                    sso_session
                )
            else:
                return self._prompt_for_registration_args_for_new_sso_session(
                    sso_session=sso_session
                )

    def _prompt_for_registration_args_with_legacy_format(self):
        self._store_sso_session_prompter_answers_to_profile_config()
        self._set_sso_session_defaults_from_profile_config()
        start_url, sso_region = self._prompt_for_sso_start_url_and_sso_region()
        return {'start_url': start_url, 'sso_region': sso_region}

    def _get_sso_registration_args_from_sso_config(self, sso_session):
        sso_config = self._get_sso_session_config(sso_session)
        return {
            'session_name': sso_session,
            'start_url': sso_config['sso_start_url'],
            'sso_region': sso_config['sso_region'],
            'registration_scopes': sso_config.get('registration_scopes'),
        }

    def _prompt_for_registration_args_for_new_sso_session(self, sso_session):
        self._set_sso_session_defaults_from_profile_config()
        start_url, sso_region = self._prompt_for_sso_start_url_and_sso_region()
        scopes = (
            self._sso_session_prompter.prompt_for_sso_registration_scopes()
        )
        return {
            'session_name': sso_session,
            'start_url': start_url,
            'sso_region': sso_region,
            'registration_scopes': scopes,
            # We force refresh for any new SSO sessions to ensure we are not
            # using any cached tokens from any previous of attempts to
            # create/authenticate a new SSO session as part of the configure
            # sso flow.
            'force_refresh': True,
        }

    def _store_sso_session_prompter_answers_to_profile_config(self):
        # Wire the SSO session prompter to set config values to the
        # dictionary used for writing to the profile section
        self._sso_session_prompter.sso_session_config = (
            self._new_profile_config_values
        )

    def _set_sso_session_in_profile_config(self, sso_session):
        self._new_profile_config_values['sso_session'] = sso_session

    def _set_sso_session_defaults_from_profile_config(self):
        # This is to ensure the SSO session prompter pulls in existing
        # SSO configuration as part of the prompt if a profile was explicitly
        # provided that already had SSO configuration
        if 'sso_start_url' in self._profile_config:
            self._sso_session_prompter.sso_session_config['sso_start_url'] = (
                self._profile_config['sso_start_url']
            )
        if 'sso_region' in self._profile_config:
            self._sso_session_prompter.sso_session_config['sso_region'] = (
                self._profile_config['sso_region']
            )

    def _prompt_for_sso_start_url_and_sso_region(self):
        start_url = self._sso_session_prompter.prompt_for_sso_start_url()
        sso_region = self._sso_session_prompter.prompt_for_sso_region()
        return start_url, sso_region

    def _warn_configuring_using_legacy_format(self):
        uni_print(
            f'{colorama.Style.BRIGHT}WARNING: Configuring using legacy format '
            f'(e.g. without an SSO session).\n'
            f'Consider re-running "configure sso" command and providing '
            f'a session name.\n{colorama.Style.RESET_ALL}'
        )

    def _prompt_for_sso_account_and_role(self, sso, sso_token):
        sso_account_id = None
        sso_role_name = None
        try:
            sso_account_id = self._prompt_for_account(sso, sso_token)
            sso_role_name = self._prompt_for_role(
                sso, sso_token, sso_account_id
            )
        except sso.exceptions.UnauthorizedException:
            uni_print(
                'Unable to list AWS accounts and/or roles. '
                'Skipping configuring AWS credential provider for profile.\n'
            )
        return sso_account_id, sso_role_name

    def _write_new_config(self, profile):
        if self._new_profile_config_values:
            profile_section = profile_to_section(profile)
            self._update_section(
                profile_section, self._new_profile_config_values
            )
        if self._sso_session_prompter.sso_session:
            self._write_sso_configuration()

    def _print_conclusion(self, configured_for_aws_credentials, profile_name):
        if configured_for_aws_credentials:
            if profile_name.lower() == 'default':
                msg = (
                    'The AWS CLI is now configured to use the default profile.\n'
                    'Run the following command to verify your configuration:\n\n'
                    'aws sts get-caller-identity\n'
                )
            else:
                msg = (
                    'To use this profile, specify the profile name using '
                    '--profile, as shown:\n\n'
                    'aws sts get-caller-identity --profile {}\n'
                )
        else:
            msg = 'Successfully configured SSO for profile: {}\n'
        uni_print(msg.format(profile_name))


class ConfigureSSOSessionCommand(BaseSSOConfigurationCommand):
    NAME = 'sso-session'
    SYNOPSIS = 'aws configure sso-session'
    DESCRIPTION = (
        'The ``aws configure sso-session`` command interactively prompts for '
        'the configuration values required to create a SSO session. '
        'The SSO session can then be associated to a profile to retrieve '
        'SSO access tokens and AWS credentials.\n\n'
        f'{_CMD_PROMPT_USAGE}'
        f'{_CONFIG_EXTRA_INFO}'
    )

    def _run_main(self, parsed_args, parsed_globals):
        self._sso_session_prompter.prompt_for_sso_session()
        self._sso_session_prompter.prompt_for_sso_start_url()
        self._sso_session_prompter.prompt_for_sso_region()
        self._sso_session_prompter.prompt_for_sso_registration_scopes()
        self._write_sso_configuration()
        self._print_configuration_success()
        return 0

    def _print_configuration_success(self):
        sso_session = self._sso_session_prompter.sso_session
        uni_print(
            f'\nCompleted configuring SSO session: {sso_session}\n'
            f'Run the following to login and refresh access token for '
            f'this session:\n\n'
            f'aws sso login --sso-session {sso_session}\n'
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


def register_ec2_page_size_injector(event_emitter):
    EC2PageSizeInjector().register(event_emitter)


class EC2PageSizeInjector:
    # Operations to auto-paginate and their specific whitelists.
    # Format:
    #    Key:   Operation
    #    Value: List of parameters to add to whitelist for that operation.
    TARGET_OPERATIONS = {
        "describe-volumes": [],
        "describe-snapshots": ['OwnerIds', 'RestorableByUserIds'],
    }

    # Parameters which should be whitelisted for every operation.
    UNIVERSAL_WHITELIST = ['NextToken', 'DryRun', 'PaginationConfig']

    DEFAULT_PAGE_SIZE = 1000

    def register(self, event_emitter):
        """Register `inject` for each target operation."""
        event_template = "calling-command.ec2.%s"
        for operation in self.TARGET_OPERATIONS:
            event = event_template % operation
            event_emitter.register_last(event, self.inject)

    def inject(self, event_name, parsed_globals, call_parameters, **kwargs):
        """Conditionally inject PageSize."""
        if not parsed_globals.paginate:
            return

        pagination_config = call_parameters.get('PaginationConfig', {})
        if 'PageSize' in pagination_config:
            return

        operation_name = event_name.split('.')[-1]

        whitelisted_params = self.TARGET_OPERATIONS.get(operation_name)
        if whitelisted_params is None:
            return

        whitelisted_params = whitelisted_params + self.UNIVERSAL_WHITELIST

        for param in call_parameters:
            if param not in whitelisted_params:
                return

        pagination_config['PageSize'] = self.DEFAULT_PAGE_SIZE
        call_parameters['PaginationConfig'] = pagination_config



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from urllib.parse import urlencode

from botocore.signers import RequestSigner


class InstanceConnectEndpointRequestSigner:
    def __init__(
        self,
        session,
        instance_connect_endpoint_dns_name,
        instance_connect_endpoint_id,
        remote_port,
        instance_private_ip,
        max_tunnel_duration,
        request_singer=None,
    ):
        service_model = session.get_service_model("ec2-instance-connect")

        if request_singer is None:
            self._request_signer = RequestSigner(
                service_model.service_id,
                session.get_config_variable("region"),
                service_model.signing_name,
                service_model.signature_version,
                session.get_credentials(),
                session.get_component("event_emitter"),
            )
        else:
            self._request_signer = request_singer

        self.instance_connect_endpoint_dns_name = (
            instance_connect_endpoint_dns_name
        )
        self.instance_connect_endpoint_id = instance_connect_endpoint_id
        self.remote_port = remote_port
        self.instance_private_ip = instance_private_ip
        self.max_tunnel_duration = max_tunnel_duration

    def get_presigned_url(self):
        qs_components = {
            "instanceConnectEndpointId": self.instance_connect_endpoint_id,
            "remotePort": self.remote_port,
            "privateIpAddress": self.instance_private_ip,
        }
        if self.max_tunnel_duration:
            qs_components["maxTunnelDuration"] = self.max_tunnel_duration
        query_string = urlencode(qs_components)

        request_dict = {
            "url": f"wss://{self.instance_connect_endpoint_dns_name}/openTunnel?{query_string}",
            "body": {},
            "headers": {"host": self.instance_connect_endpoint_dns_name},
            "method": "GET",
            "query_string": "",
            "url_path": "/",
            "context": {},
        }
        presigned_url = self._request_signer.generate_presigned_url(
            request_dict=request_dict,
            operation_name="openTunnel",
            expires_in=60,
        )
        return presigned_url



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Load wizards included in the AWS CLI.

Data Layout
===========

Within this directory, there's a ``wizards`` directory.  Each directory
maps to an existing CLI command name.

For each command directory, the filename corresponds to the name of
a wizard (with the .yml prefix stripped).  Any file that starts with
an underscore character is not listed as a wizard in the ``help`` command
output.

This means that wizards (at least these built-in ones) are only supported
from a top level CLI command.  For example::

    $ aws ec2 wizard <wizard-name>

You can also add a special file, `_main.yml` in a directory to indicate
the main wizard used when no wizard name is given.  For example::

    $ aws dynamodb wizard

will check if there's a ``dynamodb/_main.yml`` file.  If no file
exists, then a usage error will be printed.


"""

import os

from ruamel.yaml import YAML

WIZARD_SPEC_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'wizards',
)


class WizardNotExistError(Exception):
    pass


class WizardLoader:
    def __init__(self, spec_dir=None):
        if spec_dir is None:
            spec_dir = WIZARD_SPEC_DIR
        self._spec_dir = spec_dir

        self._yaml = YAML(typ='rt')

    def list_commands_with_wizards(self):
        """Returns a list of commands with at least one wizard."""
        return os.listdir(self._spec_dir)

    def list_available_wizards(self, command_name):
        """List the names of wizards available for a service.

        If no wizards are avaiable, then an empty list is returned.
        """
        files = os.listdir(os.path.join(self._spec_dir, command_name))
        return [os.path.splitext(f)[0] for f in files]

    def load_wizard(self, command_name, wizard_name):
        """Load a specific wizard.

        Given a command and wizard name (i.e something from the
        output of ``list_available_wizards``), return the loaded file.
        This will open the file and return the parsed yaml contents
        of the file.

        """
        filename = os.path.join(
            self._spec_dir, command_name, wizard_name + '.yml'
        )
        try:
            with open(filename) as f:
                return self._load_yaml(f.read())
        except OSError:
            raise WizardNotExistError(
                "Wizard does not exist for command "
                "'%s', name: '%s'" % (command_name, wizard_name)
            )

    def _load_yaml(self, contents):
        data = self._yaml.load(contents)
        return data

    def wizard_exists(self, command_name, wizard_name):
        filename = os.path.join(
            self._spec_dir, command_name, wizard_name + '.yml'
        )
        return os.path.isfile(filename)



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import base64
import binascii

from botocore.exceptions import ProfileNotFound

from awscli.shorthand import ModelVisitor


def add_binary_formatter(session, parsed_args, **kwargs):
    binary_format = parsed_args.cli_binary_format
    if binary_format is None:
        try:
            binary_format = session.get_config_variable('cli_binary_format')
        except ProfileNotFound:
            binary_format = 'base64'
    BinaryFormatHandler(binary_format).register(session)


def base64_decode_input_blobs(params, model, **kwargs):
    if model.input_shape:
        Base64DecodeVisitor().visit(params, model.input_shape)


def identity(x):
    return x


def _register_blob_parser(session, blob_parser):
    factory = session.get_component('response_parser_factory')
    factory.set_parser_defaults(blob_parser=blob_parser)


def register_identity_blob_parser(session):
    _register_blob_parser(session, identity)


class InvalidBase64Error(Exception):
    pass


class Base64DecodeVisitor(ModelVisitor):
    def _visit_scalar(self, parent, shape, name, value):
        if shape.type_name != 'blob' or not isinstance(value, str):
            return
        try:
            parent[name] = base64.b64decode(value)
        except (binascii.Error, TypeError):
            raise InvalidBase64Error('Invalid base64: "%s"' % value)


class BinaryFormatHandler:
    _BINARY_FORMATS = {
        'base64': (base64_decode_input_blobs, register_identity_blob_parser),
        'raw-in-base64-out': (None, register_identity_blob_parser),
    }

    def __init__(self, binary_format):
        self._format = binary_format

    def register(self, session):
        self._register_input_formatter(session)
        self._register_output_formatter(session)

    def _register_input_formatter(self, session):
        input_handler = self._BINARY_FORMATS[self._format][0]
        if input_handler:
            session.register('provide-client-params', input_handler)

    def _register_output_formatter(self, session):
        output_handler = self._BINARY_FORMATS[self._format][1]
        if output_handler:
            output_handler(session)
